# Copyright 2020 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains the interfaces relating to acquisition function --- functions that estimate
the utility of evaluating sets of candidate points.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Callable, Generic, Mapping, Optional

from ..data import Dataset
from ..models.interfaces import ProbabilisticModelType
from ..types import Tag, TensorType

AcquisitionFunction = Callable[[TensorType], TensorType]
"""
Type alias for acquisition functions.

An :const:`AcquisitionFunction` maps a set of `B` query points (each of dimension `D`) to a single
value that describes how useful it would be evaluate all these points together (to our goal of
optimizing the objective function). Thus, with leading dimensions, an :const:`AcquisitionFunction`
takes input shape `[..., B, D]` and returns shape `[..., 1]`.

Note that :const:`AcquisitionFunction`s which do not support batch optimization still expect inputs
with a batch dimension, i.e. an input of shape `[..., 1, D]`.
"""


class AcquisitionFunctionClass(ABC):
    """An :class:`AcquisitionFunctionClass` is an acquisition function represented using a class
    rather than as a standalone function. Using a class to represent an acquisition function
    makes it easier to update it, to avoid having to retrace the function on every call.
    """

    @abstractmethod
    def __call__(self, x: TensorType) -> TensorType:
        """Call acquisition function."""


class AcquisitionFunctionBuilder(Generic[ProbabilisticModelType], ABC):
    """An :class:`AcquisitionFunctionBuilder` builds and updates an acquisition function."""

    @abstractmethod
    def prepare_acquisition_function(
        self,
        models: Mapping[Tag, ProbabilisticModelType],
        datasets: Optional[Mapping[Tag, Dataset]] = None,
    ) -> AcquisitionFunction:
        """
        Prepare an acquisition function. We assume that this requires at least models, but
        it may sometimes also need data.

        :param models: The models for each tag.
        :param datasets: The data from the observer (optional).
        :return: An acquisition function.
        """

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        models: Mapping[Tag, ProbabilisticModelType],
        datasets: Optional[Mapping[Tag, Dataset]] = None,
    ) -> AcquisitionFunction:
        """
        Update an acquisition function. By default this generates a new acquisition function each
        time. However, if the function is decorated with `@tf.function`, then you can override
        this method to update its variables instead and avoid retracing the acquisition function on
        every optimization loop.

        :param function: The acquisition function to update.
        :param models: The models for each tag.
        :param datasets: The data from the observer (optional).
        :return: The updated acquisition function.
        """
        return self.prepare_acquisition_function(models, datasets=datasets)


class SingleModelAcquisitionBuilder(Generic[ProbabilisticModelType], ABC):
    """
    Convenience acquisition function builder for an acquisition function (or component of a
    composite acquisition function) that requires only one model, dataset pair.
    """

    def using(self, tag: Tag) -> AcquisitionFunctionBuilder[ProbabilisticModelType]:
        """
        :param tag: The tag for the model, dataset pair to use to build this acquisition function.
        :return: An acquisition function builder that selects the model and dataset specified by
            ``tag``, as defined in :meth:`prepare_acquisition_function`.
        """

        class _Anon(AcquisitionFunctionBuilder[ProbabilisticModelType]):
            def __init__(
                self, single_builder: SingleModelAcquisitionBuilder[ProbabilisticModelType]
            ):
                self.single_builder = single_builder

            def prepare_acquisition_function(
                self,
                models: Mapping[Tag, ProbabilisticModelType],
                datasets: Optional[Mapping[Tag, Dataset]] = None,
            ) -> AcquisitionFunction:
                return self.single_builder.prepare_acquisition_function(
                    models[tag], dataset=None if datasets is None else datasets[tag]
                )

            def update_acquisition_function(
                self,
                function: AcquisitionFunction,
                models: Mapping[Tag, ProbabilisticModelType],
                datasets: Optional[Mapping[Tag, Dataset]] = None,
            ) -> AcquisitionFunction:
                return self.single_builder.update_acquisition_function(
                    function, models[tag], dataset=None if datasets is None else datasets[tag]
                )

            def __repr__(self) -> str:
                return f"{self.single_builder!r} using tag {tag!r}"

        return _Anon(self)

    @abstractmethod
    def prepare_acquisition_function(
        self,
        model: ProbabilisticModelType,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The model.
        :param dataset: The data to use to build the acquisition function (optional).
        :return: An acquisition function.
        """

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: ProbabilisticModelType,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model.
        :param dataset: The data from the observer (optional).
        :return: The updated acquisition function.
        """
        return self.prepare_acquisition_function(model, dataset=dataset)


class GreedyAcquisitionFunctionBuilder(Generic[ProbabilisticModelType], ABC):
    """
    A :class:`GreedyAcquisitionFunctionBuilder` builds an acquisition function
    suitable for greedily building batches for batch Bayesian
    Optimization. A :class:`GreedyAcquisitionFunctionBuilder` differs
    from an :class:`AcquisitionFunctionBuilder` by requiring that a set
    of pending points is passed to the builder. Note that this acquisition function
    is typically called `B` times each Bayesian optimization step, when building batches
    of size `B`.
    """

    @abstractmethod
    def prepare_acquisition_function(
        self,
        models: Mapping[Tag, ProbabilisticModelType],
        datasets: Optional[Mapping[Tag, Dataset]] = None,
        pending_points: Optional[TensorType] = None,
    ) -> AcquisitionFunction:
        """
        Generate a new acquisition function. The first time this is called, ``pending_points``
        will be `None`. Subsequent calls will be via ``update_acquisition_function`` below,
        unless that has been overridden.

        :param models: The models over each tag.
        :param datasets: The data from the observer (optional).
        :param pending_points: Points already chosen to be in the current batch (of shape [M,D]),
            where M is the number of pending points and D is the search space dimension.
        :return: An acquisition function.
        """

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        models: Mapping[Tag, ProbabilisticModelType],
        datasets: Optional[Mapping[Tag, Dataset]] = None,
        pending_points: Optional[TensorType] = None,
        new_optimization_step: bool = True,
    ) -> AcquisitionFunction:
        """
        Update an acquisition function. By default this generates a new acquisition function each
        time. However, if the function is decorated with`@tf.function`, then you can override
        this method to update its variables instead and avoid retracing the acquisition function on
        every optimization loop.

        :param function: The acquisition function to update.
        :param models: The models over each tag.
        :param datasets: The data from the observer (optional).
        :param pending_points: Points already chosen to be in the current batch (of shape [M,D]),
            where M is the number of pending points and D is the search space dimension.
        :param new_optimization_step: Indicates whether this call to update_acquisition_function
            is to start of a new optimization step, of to continue collecting batch of points
            for the current step. Defaults to ``True``.
        :return: The updated acquisition function.
        """
        return self.prepare_acquisition_function(
            models, datasets=datasets, pending_points=pending_points
        )


class SingleModelGreedyAcquisitionBuilder(Generic[ProbabilisticModelType], ABC):
    """
    Convenience acquisition function builder for a greedy acquisition function (or component of a
    composite greedy acquisition function) that requires only one model, dataset pair.
    """

    def using(self, tag: Tag) -> GreedyAcquisitionFunctionBuilder[ProbabilisticModelType]:
        """
        :param tag: The tag for the model, dataset pair to use to build this acquisition function.
        :return: An acquisition function builder that selects the model and dataset specified by
            ``tag``, as defined in :meth:`prepare_acquisition_function`.
        """

        class _Anon(GreedyAcquisitionFunctionBuilder[ProbabilisticModelType]):
            def __init__(
                self, single_builder: SingleModelGreedyAcquisitionBuilder[ProbabilisticModelType]
            ):
                self.single_builder = single_builder

            def prepare_acquisition_function(
                self,
                models: Mapping[Tag, ProbabilisticModelType],
                datasets: Optional[Mapping[Tag, Dataset]] = None,
                pending_points: Optional[TensorType] = None,
            ) -> AcquisitionFunction:
                return self.single_builder.prepare_acquisition_function(
                    models[tag],
                    dataset=None if datasets is None else datasets[tag],
                    pending_points=pending_points,
                )

            def update_acquisition_function(
                self,
                function: AcquisitionFunction,
                models: Mapping[Tag, ProbabilisticModelType],
                datasets: Optional[Mapping[Tag, Dataset]] = None,
                pending_points: Optional[TensorType] = None,
                new_optimization_step: bool = True,
            ) -> AcquisitionFunction:
                return self.single_builder.update_acquisition_function(
                    function,
                    models[tag],
                    dataset=None if datasets is None else datasets[tag],
                    pending_points=pending_points,
                    new_optimization_step=new_optimization_step,
                )

            def __repr__(self) -> str:
                return f"{self.single_builder!r} using tag {tag!r}"

        return _Anon(self)

    @abstractmethod
    def prepare_acquisition_function(
        self,
        model: ProbabilisticModelType,
        dataset: Optional[Dataset] = None,
        pending_points: Optional[TensorType] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The model.
        :param dataset: The data from the observer (optional).
        :param pending_points: Points already chosen to be in the current batch (of shape [M,D]),
            where M is the number of pending points and D is the search space dimension.
        :return: An acquisition function.
        """

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: ProbabilisticModelType,
        dataset: Optional[Dataset] = None,
        pending_points: Optional[TensorType] = None,
        new_optimization_step: bool = True,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model.
        :param dataset: The data from the observer (optional).
        :param pending_points: Points already chosen to be in the current batch (of shape [M,D]),
            where M is the number of pending points and D is the search space dimension.
        :param new_optimization_step: Indicates whether this call to update_acquisition_function
            is to start of a new optimization step, of to continue collecting batch of points
            for the current step. Defaults to ``True``.
        :return: The updated acquisition function.
        """
        return self.prepare_acquisition_function(
            model,
            dataset=dataset,
            pending_points=pending_points,
        )


class VectorizedAcquisitionFunctionBuilder(AcquisitionFunctionBuilder[ProbabilisticModelType]):
    """
    An :class:`VectorizedAcquisitionFunctionBuilder` builds and updates a vectorized
    acquisition function These differ from normal acquisition functions only by their output shape:
    rather than returning a single value, they return one value per potential query point.
    Thus, with leading dimensions, they take input shape `[..., B, D]` and returns shape `[..., B]`.
    """


class SingleModelVectorizedAcquisitionBuilder(
    SingleModelAcquisitionBuilder[ProbabilisticModelType]
):
    """
    Convenience acquisition function builder for vectorized acquisition functions (or component
    of a composite vectorized acquisition function) that requires only one model, dataset pair.
    """

    def using(self, tag: Tag) -> AcquisitionFunctionBuilder[ProbabilisticModelType]:
        """
        :param tag: The tag for the model, dataset pair to use to build this acquisition function.
        :return: An acquisition function builder that selects the model and dataset specified by
            ``tag``, as defined in :meth:`prepare_acquisition_function`.
        """

        class _Anon(VectorizedAcquisitionFunctionBuilder[ProbabilisticModelType]):
            def __init__(
                self,
                single_builder: SingleModelVectorizedAcquisitionBuilder[ProbabilisticModelType],
            ):
                self.single_builder = single_builder

            def prepare_acquisition_function(
                self,
                models: Mapping[Tag, ProbabilisticModelType],
                datasets: Optional[Mapping[Tag, Dataset]] = None,
            ) -> AcquisitionFunction:
                return self.single_builder.prepare_acquisition_function(
                    models[tag], dataset=None if datasets is None else datasets[tag]
                )

            def update_acquisition_function(
                self,
                function: AcquisitionFunction,
                models: Mapping[Tag, ProbabilisticModelType],
                datasets: Optional[Mapping[Tag, Dataset]] = None,
            ) -> AcquisitionFunction:
                return self.single_builder.update_acquisition_function(
                    function, models[tag], dataset=None if datasets is None else datasets[tag]
                )

            def __repr__(self) -> str:
                return f"{self.single_builder!r} using tag {tag!r}"

        return _Anon(self)


PenalizationFunction = Callable[[TensorType], TensorType]
"""
An :const:`PenalizationFunction` maps a query point (of dimension `D`) to a single
value that described how heavily it should be penalized (a positive quantity).
As penalization is applied multiplicatively to acquisition functions, small
penalization outputs correspond to a stronger penalization effect. Thus, with
leading dimensions, an :const:`PenalizationFunction` takes input
shape `[..., 1, D]` and returns shape `[..., 1]`.
"""


class UpdatablePenalizationFunction(ABC):
    """An :class:`UpdatablePenalizationFunction` builds and updates a penalization function.
    Defining a penalization function that can be updated avoids having to retrace on every call."""

    @abstractmethod
    def __call__(self, x: TensorType) -> TensorType:
        """Call penalization function.."""

    @abstractmethod
    def update(
        self,
        pending_points: TensorType,
        lipschitz_constant: TensorType,
        eta: TensorType,
    ) -> None:
        """Update penalization function."""



# Copyright 2020 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" This module contains implementations of various types of search space. """
from __future__ import annotations

import operator
from abc import ABC, abstractmethod
from functools import reduce
from itertools import chain
from typing import Callable, Optional, Sequence, Tuple, TypeVar, Union, overload

import numpy as np
import scipy.optimize as spo
import tensorflow as tf
import tensorflow_probability as tfp
from check_shapes import check_shapes
from gpflow.keras import tf_keras
from typing_extensions import Protocol, runtime_checkable

from .types import TensorType
from .utils import flatten_leading_dims

SearchSpaceType = TypeVar("SearchSpaceType", bound="SearchSpace")
""" A type variable bound to :class:`SearchSpace`. """

DEFAULT_DTYPE: tf.DType = tf.float64
""" Default dtype to use when none is provided. """

EncoderFunction = Callable[[TensorType], TensorType]
""" Type alias for point encoders. These transform points from one search space to another. """


class SampleTimeoutError(Exception):
    """Raised when sampling from a search space has timed out."""


class NonlinearConstraint(spo.NonlinearConstraint):  # type: ignore[misc]
    """
    A wrapper class for nonlinear constraints on variables. The constraints expression is of the
    form::

        lb <= fun(x) <= ub

    :param fun: The function defining the nonlinear constraints; with input shape [..., D] and
        output shape [..., 1], returning a scalar value for each input point.
    :param lb: The lower bound of the constraint. Should be a scalar or of shape [1].
    :param ub: The upper bound of the constraint. Should be a scalar or of shape [1].
    :param keep_feasible: Keep the constraints feasible throughout optimization iterations if this
        is `True`.
    """

    def __init__(
        self,
        fun: Callable[[TensorType], TensorType],
        lb: Sequence[float] | TensorType,
        ub: Sequence[float] | TensorType,
        keep_feasible: bool = False,
    ):
        # Implement caching to avoid calling the constraint function multiple times to get value
        # and gradient.
        def _constraint_value_and_gradient(x: TensorType) -> Tuple[TensorType, TensorType]:
            val, grad = tfp.math.value_and_gradient(fun, x)

            tf.debugging.assert_shapes(
                [(val, [..., 1])],
                message="Nonlinear constraint only supports single output function.",
            )

            return tf.cast(val, dtype=x.dtype), tf.cast(grad, dtype=x.dtype)

        cache_x: TensorType = tf.constant([])
        cache_f: TensorType = tf.constant([])
        cache_df_dx: TensorType = tf.constant([])

        def val_fun(x: TensorType) -> TensorType:
            nonlocal cache_x, cache_f, cache_df_dx
            if not np.array_equal(x, cache_x):
                cache_f, cache_df_dx = _constraint_value_and_gradient(x)
                cache_x = x
            return cache_f

        def jac_fun(x: TensorType) -> TensorType:
            nonlocal cache_x, cache_f, cache_df_dx
            if not np.array_equal(x, cache_x):
                cache_f, cache_df_dx = _constraint_value_and_gradient(x)
                cache_x = x
            return cache_df_dx

        self._orig_fun = fun  # Used for constraints equality check.
        super().__init__(val_fun, lb, ub, jac=jac_fun, keep_feasible=keep_feasible)

    def residual(self, points: TensorType) -> TensorType:
        """
        Calculate the residuals between the constraint function and its lower/upper limits.

        :param points: The points to calculate the residuals for, with shape [..., D].
        :return: A tensor containing the lower and upper residual values with shape [..., 2].
        """
        tf.debugging.assert_rank_at_least(points, 2)
        non_d_axes = np.ones_like(points.shape)[:-1]  # Avoid adding axes shape to static graph.
        lb = tf.cast(tf.reshape(self.lb, (*non_d_axes, -1)), dtype=points.dtype)
        ub = tf.cast(tf.reshape(self.ub, (*non_d_axes, -1)), dtype=points.dtype)
        fval = self.fun(points)
        fval = tf.reshape(fval, (*points.shape[:-1], -1))  # Atleast 2D.
        fval = tf.cast(fval, dtype=points.dtype)
        values = [fval - lb, ub - fval]
        values = tf.concat(values, axis=-1)
        return values

    def __repr__(self) -> str:
        """"""
        return f"""
            NonlinearConstraint({self.fun!r}, {self.lb!r}, {self.ub!r}, {self.keep_feasible!r})"
        """

    def __eq__(self, other: object) -> bool:
        """
        :param other: A constraint.
        :return: Whether the constraint is identical to this one.
        """
        if not isinstance(other, NonlinearConstraint):
            return False
        return bool(
            self._orig_fun == other._orig_fun
            and tf.reduce_all(self.lb == other.lb)
            and tf.reduce_all(self.ub == other.ub)
            and self.keep_feasible == other.keep_feasible
        )


class LinearConstraint(spo.LinearConstraint):  # type: ignore[misc]
    """
    A wrapper class for linear constraints on variables. The constraints expression is of the form::

        lb <= A @ x <= ub

    :param A: The matrix defining the linear constraints with shape [M, D], where M is the
        number of constraints.
    :param lb: The lower bound of the constraint. Should be a scalar or of shape [M].
    :param ub: The upper bound of the constraint. Should be a scalar or of shape [M].
    :param keep_feasible: Keep the constraints feasible throughout optimization iterations if this
        is `True`.
    """

    def __init__(
        self,
        A: TensorType,
        lb: Sequence[float] | TensorType,
        ub: Sequence[float] | TensorType,
        keep_feasible: bool = False,
    ):
        super().__init__(A, lb, ub, keep_feasible=keep_feasible)

    def residual(self, points: TensorType) -> TensorType:
        """
        Calculate the residuals between the constraint function and its lower/upper limits.

        :param points: The points to calculate the residuals for, with shape [..., D].
        :return: A tensor containing the lower and upper residual values with shape [..., M*2].
        """
        tf.debugging.assert_rank_at_least(points, 2)
        non_d_axes = np.ones_like(points.shape)[:-1]  # Avoid adding axes shape to static graph.
        lb = tf.cast(tf.reshape(self.lb, (*non_d_axes, -1)), dtype=points.dtype)
        ub = tf.cast(tf.reshape(self.ub, (*non_d_axes, -1)), dtype=points.dtype)
        A = tf.cast(self.A, dtype=points.dtype)
        fval = tf.linalg.matmul(points, A, transpose_b=True)
        fval = tf.reshape(fval, (*points.shape[:-1], -1))  # Atleast 2D.
        values = [fval - lb, ub - fval]
        values = tf.concat(values, axis=-1)
        return values

    def __repr__(self) -> str:
        """"""
        return f"""
            LinearConstraint({self.A!r}, {self.lb!r}, {self.ub!r}, {self.keep_feasible!r})"
        """

    def __eq__(self, other: object) -> bool:
        """
        :param other: A constraint.
        :return: Whether the constraint is identical to this one.
        """
        if not isinstance(other, LinearConstraint):
            return False
        return bool(
            tf.reduce_all(self.A == other.A)
            and tf.reduce_all(self.lb == other.lb)
            and tf.reduce_all(self.ub == other.ub)
            and tf.reduce_all(self.keep_feasible == other.keep_feasible)
        )


Constraint = Union[LinearConstraint, NonlinearConstraint]
""" Type alias for constraints. """


class SearchSpace(ABC):
    """
    A :class:`SearchSpace` represents the domain over which an objective function is optimized.
    """

    @abstractmethod
    def sample(self, num_samples: int, seed: Optional[int] = None) -> TensorType:
        """
        :param num_samples: The number of points to sample from this search space.
        :param seed: Random seed for reproducibility.
        :return: ``num_samples`` i.i.d. random points, sampled uniformly from this search space.
        """

    def contains(self, value: TensorType) -> TensorType:
        """Method for checking membership.

        :param value: A point or points to check for membership of this :class:`SearchSpace`.
        :return: A boolean array showing membership for each point in value.
        :raise ValueError (or tf.errors.InvalidArgumentError): If ``value`` has a different
            dimensionality points from this :class:`SearchSpace`.
        """
        tf.debugging.assert_equal(
            tf.rank(value) > 0 and tf.shape(value)[-1] == self.dimension,
            True,
            message=f"""
                Dimensionality mismatch: space is {self.dimension}, value is {tf.shape(value)[-1]}
                """,
        )
        return self._contains(value)

    @abstractmethod
    def _contains(self, value: TensorType) -> TensorType:
        """Space-specific implementation of membership. Can assume valid input shape.

        :param value: A point or points to check for membership of this :class:`SearchSpace`.
        :return: A boolean array showing membership for each point in value.
        """

    def __contains__(self, value: TensorType) -> bool:
        """Method called by `in` operator. Doesn't support broadcasting as Python insists
        on converting the result to a boolean.

        :param value: A single point to check for membership of this :class:`SearchSpace`.
        :return: `True` if ``value`` is a member of this search space, else `False`.
        :raise ValueError (or tf.errors.InvalidArgumentError): If ``value`` has a different
            dimensionality from this :class:`SearchSpace`.
        """
        tf.debugging.assert_equal(
            tf.rank(value) == 1,
            True,
            message=f"""
                Rank mismatch: expected 1, got {tf.rank(value)}. To get a tensor of boolean
                membership values from a tensor of points, use `space.contains(value)`
                rather than `value in space`.
                """,
        )
        return self.contains(value)

    @property
    @abstractmethod
    def dimension(self) -> TensorType:
        """The number of inputs in this search space."""

    @property
    @abstractmethod
    def has_bounds(self) -> bool:
        """Whether the search space has meaningful numerical bounds."""

    @property
    @abstractmethod
    def lower(self) -> TensorType:
        """The lowest value taken by each search space dimension."""

    @property
    @abstractmethod
    def upper(self) -> TensorType:
        """The highest value taken by each search space dimension."""

    @abstractmethod
    def product(self: SearchSpaceType, other: SearchSpaceType) -> SearchSpaceType:
        """
        :param other: A search space of the same type as this search space.
        :return: The Cartesian product of this search space with the ``other``.
        """

    @overload
    def __mul__(self: SearchSpaceType, other: SearchSpaceType) -> SearchSpaceType: ...

    @overload
    def __mul__(self: SearchSpaceType, other: SearchSpace) -> SearchSpace:  # type: ignore[misc]
        # mypy complains that this is superfluous, but it seems to use it fine to infer
        # that Box * Box = Box, while Box * Discrete = SearchSpace.
        ...

    def __mul__(self, other: SearchSpace) -> SearchSpace:
        """
        :param other: A search space.
        :return: The Cartesian product of this search space with the ``other``.
            If both spaces are of the same type then this calls the :meth:`product` method.
            Otherwise, it generates a :class:`TaggedProductSearchSpace`.
        """
        # If the search space has any constraints, always return a tagged product search space.
        if not self.has_constraints and not other.has_constraints and isinstance(other, type(self)):
            return self.product(other)
        return TaggedProductSearchSpace((self, other))

    def __pow__(self: SearchSpaceType, other: int) -> SearchSpaceType:
        """
        Return the Cartesian product of ``other`` instances of this search space. For example, for
        an exponent of `3`, and search space `s`, this is `s ** 3`, which is equivalent to
        `s * s * s`.

        :param other: The exponent, or number of instances of this search space to multiply
            together. Must be strictly positive.
        :return: The Cartesian product of ``other`` instances of this search space.
        :raise tf.errors.InvalidArgumentError: If the exponent ``other`` is less than 1.
        """
        tf.debugging.assert_positive(other, message="Exponent must be strictly positive")
        return reduce(operator.mul, [self] * other)

    def discretize(self, num_samples: int) -> DiscreteSearchSpace:
        """
        :param num_samples: The number of points in the :class:`DiscreteSearchSpace`.
        :return: A discrete search space consisting of ``num_samples`` points sampled uniformly from
            this search space.
        :raise NotImplementedError: If this :class:`SearchSpace` has constraints.
        """
        if self.has_constraints:  # Constraints are not supported.
            raise NotImplementedError(
                "Discretization is currently not supported in the presence of constraints."
            )
        return DiscreteSearchSpace(points=self.sample(num_samples))

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        """
        :param other: A search space.
        :return: Whether the search space is identical to this one.
        """

    @property
    def constraints(self) -> Sequence[Constraint]:
        """The sequence of explicit constraints specified in this search space."""
        return []

    def constraints_residuals(self, points: TensorType) -> TensorType:
        """
        Return residuals for all the constraints in this :class:`SearchSpace`.

        :param points: The points to get the residuals for, with shape [..., D].
        :return: A tensor of all the residuals with shape [..., C], where C is the total number of
            constraints.
        :raise NotImplementedError: If this :class:`SearchSpace` does not support constraints.
        """
        raise NotImplementedError("Constraints are currently not supported for this search space.")

    def is_feasible(self, points: TensorType) -> TensorType:
        """
        Checks if points satisfy the explicit constraints of this :class:`SearchSpace`.
        Note membership of the search space is not checked.

        :param points: The points to check constraints feasibility for, with shape [..., D].
        :return: A tensor of booleans. Returns `True` for each point if it is feasible in this
            search space, else `False`.
        :raise NotImplementedError: If this :class:`SearchSpace` has constraints.
        """
        # Everything is feasible in the absence of constraints. Must be overriden if there are
        # constraints.
        if self.has_constraints:
            raise NotImplementedError("Feasibility check is not implemented for this search space.")
        return tf.cast(tf.ones(points.shape[:-1]), dtype=bool)

    @property
    def has_constraints(self) -> bool:
        """Returns `True` if this search space has any explicit constraints specified."""
        # By default, assume there are no constraints; can be overridden by a subclass.
        return False


class GeneralDiscreteSearchSpace(SearchSpace):
    """
    An ABC representing different types of discrete search spaces (not just numerical).
    This contains a default implementation using explicitly provided points which subclasses
    may ignore.
    """

    def __init__(self, points: TensorType):
        """
        :param points: The points that define the discrete space, with shape ('N', 'D').
        :raise ValueError (or tf.errors.InvalidArgumentError): If ``points`` has an invalid shape.
        """

        tf.debugging.assert_shapes([(points, ("N", "D"))])
        self._points = points
        self._dimension = tf.shape(self._points)[-1]

    @property
    def points(self) -> TensorType:
        """All the points in this space."""
        return self._points

    @property
    def dimension(self) -> TensorType:
        """The number of inputs in this search space."""
        return self._dimension

    def _contains(self, value: TensorType) -> TensorType:
        comparison = tf.math.equal(self.points, tf.expand_dims(value, -2))  # [..., N, D]
        return tf.reduce_any(tf.reduce_all(comparison, axis=-1), axis=-1)  # [...]

    def sample(self, num_samples: int, seed: Optional[int] = None) -> TensorType:
        """
        :param num_samples: The number of points to sample from this search space.
        :param seed: Random seed for reproducibility.
        :return: ``num_samples`` i.i.d. random points, sampled uniformly,
            from this search space.
        """
        if seed is not None:  # ensure reproducibility
            tf.random.set_seed(seed)

        if num_samples == 0:
            return self.points[:0, :]
        else:
            sampled_indices = tf.random.categorical(
                tf.ones((1, tf.shape(self.points)[0])), num_samples, seed=seed
            )
            return tf.gather(self.points, sampled_indices)[0, :, :]  # [num_samples, D]


class DiscreteSearchSpace(GeneralDiscreteSearchSpace):
    r"""
    A discrete :class:`SearchSpace` representing a finite set of :math:`D`-dimensional points in
    :math:`\mathbb{R}^D`.

    For example:

        >>> points = tf.constant([[-1.0, 0.4], [-1.0, 0.6], [0.0, 0.4]])
        >>> search_space = DiscreteSearchSpace(points)
        >>> assert tf.constant([0.0, 0.4]) in search_space
        >>> assert tf.constant([1.0, 0.5]) not in search_space

    """

    def __repr__(self) -> str:
        """"""
        return f"DiscreteSearchSpace({self._points!r})"

    @property
    def has_bounds(self) -> bool:
        return True

    @property
    def lower(self) -> TensorType:
        """The lowest value taken across all points by each search space dimension."""
        return tf.reduce_min(self.points, -2)

    @property
    def upper(self) -> TensorType:
        """The highest value taken across all points by each search space dimension."""
        return tf.reduce_max(self.points, -2)

    def product(self, other: DiscreteSearchSpace) -> DiscreteSearchSpace:
        r"""
        Return the Cartesian product of the two :class:`DiscreteSearchSpace`\ s. For example:

            >>> sa = DiscreteSearchSpace(tf.constant([[0, 1], [2, 3]]))
            >>> sb = DiscreteSearchSpace(tf.constant([[4, 5, 6], [7, 8, 9]]))
            >>> (sa * sb).points.numpy()
            array([[0, 1, 4, 5, 6],
                   [0, 1, 7, 8, 9],
                   [2, 3, 4, 5, 6],
                   [2, 3, 7, 8, 9]], dtype=int32)

        :param other: A :class:`DiscreteSearchSpace` with :attr:`points` of the same dtype as this
            search space.
        :return: The Cartesian product of the two :class:`DiscreteSearchSpace`\ s.
        :raise TypeError: If one :class:`DiscreteSearchSpace` has :attr:`points` of a different
            dtype to the other.
        """
        if self.points.dtype is not other.points.dtype:
            return NotImplemented

        tile_self = tf.tile(self.points[:, None], [1, len(other.points), 1])
        tile_other = tf.tile(other.points[None], [len(self.points), 1, 1])
        cartesian_product = tf.concat([tile_self, tile_other], axis=2)
        product_space_dimension = self.points.shape[-1] + other.points.shape[-1]
        return DiscreteSearchSpace(tf.reshape(cartesian_product, [-1, product_space_dimension]))

    def __eq__(self, other: object) -> bool:
        """
        :param other: A search space.
        :return: Whether the search space is identical to this one.
        """
        if not isinstance(other, DiscreteSearchSpace):
            return NotImplemented
        return bool(tf.reduce_all(tf.sort(self.points, 0) == tf.sort(other.points, 0)))


@runtime_checkable
class HasOneHotEncoder(Protocol):
    """A categorical search space that contains default logic for one-hot encoding."""

    @property
    @abstractmethod
    def one_hot_encoder(self) -> EncoderFunction:
        "A one-hot encoder for points in the search space."


def one_hot_encoder(space: SearchSpace) -> EncoderFunction:
    "A utility function for one-hot encoding a search space when it supports it."
    return space.one_hot_encoder if isinstance(space, HasOneHotEncoder) else lambda x: x


def cast_encoder(
    encoder: EncoderFunction,
    input_dtype: Optional[tf.DType] = None,
    output_dtype: Optional[tf.DType] = None,
) -> EncoderFunction:
    "A utility function for casting the input and/or output of an encoder."

    def cast_and_encode(x: TensorType) -> TensorType:
        if input_dtype is not None:
            x = tf.cast(x, input_dtype)
        y = encoder(x)
        if output_dtype is not None:
            y = tf.cast(y, output_dtype)
        return y

    return cast_and_encode


def one_hot_encoded_space(space: SearchSpace) -> SearchSpace:
    "A bounded search space corresponding to the one-hot encoding of the given space."

    if isinstance(space, GeneralDiscreteSearchSpace) and isinstance(space, HasOneHotEncoder):
        return DiscreteSearchSpace(space.one_hot_encoder(space.points))
    elif isinstance(space, TaggedProductSearchSpace):
        spaces = [one_hot_encoded_space(space.get_subspace(tag)) for tag in space.subspace_tags]
        return TaggedProductSearchSpace(spaces=spaces, tags=space.subspace_tags)
    elif isinstance(space, HasOneHotEncoder):
        raise NotImplementedError(f"Unsupported one-hot-encoded space {type(space)}")
    else:
        return space


class CategoricalSearchSpace(GeneralDiscreteSearchSpace, HasOneHotEncoder):
    r"""
    A categorical :class:`SearchSpace` representing a finite set :math:`\mathcal{C}` of categories,
    or a finite Cartesian product :math:`\mathcal{C}_1 \times \cdots \times \mathcal{C}_n` of
    such sets.

    For example:

        >>> CategoricalSearchSpace(5)
        CategoricalSearchSpace([('0', '1', '2', '3', '4')])
        >>> CategoricalSearchSpace(["Red", "Green", "Blue"])
        CategoricalSearchSpace([('Red', 'Green', 'Blue')])
        >>> CategoricalSearchSpace([2,3])
        CategoricalSearchSpace([('0', '1'), ('0', '1', '2')])
        >>> CategoricalSearchSpace([["R", "G", "B"], ["Y", "N"]])
        CategoricalSearchSpace([('R', 'G', 'B'), ('Y', 'N')])

    Note that internally categories are represented by numeric indices:

        >>> rgb = CategoricalSearchSpace(["Red", "Green", "Blue"])
        >>> assert tf.constant([1], dtype=tf.float64) in rgb
        >>> assert tf.constant([3], dtype=tf.float64) not in rgb
        >>> rgb.to_tags(tf.constant([[1], [0], [2]]))
        <tf.Tensor: shape=(3, 1), dtype=string, numpy=
        array([[b'Green'],
               [b'Red'],
               [b'Blue']], dtype=object)>

    """

    def __init__(
        self,
        categories: int | Sequence[int] | Sequence[str] | Sequence[Sequence[str]],
        dtype: tf.DType = DEFAULT_DTYPE,
    ):
        """
        :param categories: Number of categories or category names. Can be an array for
            multidimensional spaces.
        :param dtype: The dtype of the returned indices, either tf.float32 or tf.float64.
        """
        if isinstance(categories, int) or any(isinstance(x, str) for x in categories):
            categories = [categories]  # type: ignore[assignment]

        if not isinstance(categories, Sequence) or not (
            all(
                isinstance(x, Sequence)
                and not isinstance(x, str)
                and all(isinstance(y, str) for y in x)
                for x in categories
            )
            or all(isinstance(x, int) for x in categories)
        ):
            raise TypeError(
                f"Invalid category description {categories!r}: " "expected either numbers or names."
            )

        elif any(isinstance(x, int) for x in categories):
            category_lens: Sequence[int] = categories  # type: ignore[assignment]
            if any(x <= 0 for x in category_lens):
                raise ValueError("Numbers of categories must be positive")
            tags = [tuple(f"{i}" for i in range(n)) for n in category_lens]
        else:
            category_names: Sequence[Sequence[str]] = categories  # type: ignore[assignment]
            if any(len(ts) == 0 for ts in category_names):
                raise ValueError("Category name lists cannot be empty")
            tags = [tuple(ts) for ts in category_names]

        self._tags = tags
        self._dtype = dtype

        ranges = [tf.range(len(ts), dtype=dtype) for ts in tags]
        meshgrid = tf.meshgrid(*ranges, indexing="ij")
        points = (
            tf.reshape(tf.stack(meshgrid, axis=-1), [-1, len(tags)]) if tags else tf.zeros([0, 0])
        )

        super().__init__(points)

    def __repr__(self) -> str:
        """"""
        return f"CategoricalSearchSpace({self._tags!r})"

    @property
    def has_bounds(self) -> bool:
        return False

    @property
    def lower(self) -> TensorType:
        raise AttributeError("Categorical search spaces do not have numerical bounds")

    @property
    def upper(self) -> TensorType:
        raise AttributeError("Categorical search spaces do not have numerical bounds")

    @property
    def tags(self) -> Sequence[Sequence[str]]:
        """The tags of the categories."""
        return self._tags

    @property
    def one_hot_encoder(self) -> EncoderFunction:
        """A one-hot encoder for the numerical indices. Note that binary categories
        are left unchanged instead of adding an unnecessary second feature."""

        def binary_encoder(x: TensorType) -> TensorType:
            # no need to one-hot encode binary categories (but we should still validate)
            tf.debugging.Assert(tf.reduce_all((x == 0) | (x == 1)), [tf.constant([])])
            return x

        def encoder(x: TensorType) -> TensorType:
            flat_x, unflatten = flatten_leading_dims(x)
            tf.debugging.assert_equal(flat_x.shape[-1], len(self.tags))
            columns = tf.split(flat_x, flat_x.shape[-1], axis=1)
            encoders = [
                (
                    binary_encoder
                    if len(ts) == 2
                    else tf_keras.layers.CategoryEncoding(num_tokens=len(ts), output_mode="one_hot")
                )
                for ts in self.tags
            ]
            encoded = tf.concat(
                [
                    tf.cast(encoder(column), dtype=self._dtype)
                    for encoder, column in zip(encoders, columns)
                ],
                axis=1,
            )
            return unflatten(encoded)

        return encoder

    def to_tags(self, indices: TensorType) -> TensorType:
        """
        Convert a tensor of indices (such as one returned by :meth:`sample`) to one of
        category tags.

        :param indices: A tensor of integer indices.
        :return: A tensor of string tags.
        """
        if indices.dtype.is_floating:
            if not tf.reduce_all(tf.math.equal(indices, tf.math.floor(indices))):
                raise ValueError("Non-integral indices passed to to_tags")
            indices = tf.cast(indices, dtype=tf.int32)

        def extract_tags(row: TensorType) -> TensorType:
            return tf.stack(
                [tf.gather(tf.constant(self._tags[i]), row[i]) for i in range(len(row))]
            )

        return tf.map_fn(extract_tags, indices, dtype=tf.string)

    def product(self, other: CategoricalSearchSpace) -> CategoricalSearchSpace:
        r"""
        Return the Cartesian product of the two :class:`CategoricalSearchSpace`\ s. For example:

            >>> rgb = CategoricalSearchSpace(["Red", "Green", "Blue"])
            >>> yn = CategoricalSearchSpace(["Yes", "No"])
            >>> rgb * yn
            CategoricalSearchSpace([('Red', 'Green', 'Blue'), ('Yes', 'No')])

        :param other: A :class:`CategoricalSearchSpace`.
        :return: The Cartesian product of the two :class:`CategoricalSearchSpace`\ s.
        """
        return CategoricalSearchSpace(tuple(chain(self.tags, other.tags)))

    def __eq__(self, other: object) -> bool:
        """
        :param other: A search space.
        :return: Whether the search space is identical to this one.
        """
        if not isinstance(other, CategoricalSearchSpace):
            return NotImplemented
        return self.tags == other.tags


class Box(SearchSpace):
    r"""
    Continuous :class:`SearchSpace` representing a :math:`D`-dimensional box in
    :math:`\mathbb{R}^D`. Mathematically it is equivalent to the Cartesian product of :math:`D`
    closed bounded intervals in :math:`\mathbb{R}`.
    """

    @overload
    def __init__(
        self,
        lower: Sequence[float],
        upper: Sequence[float],
        constraints: Optional[Sequence[Constraint]] = None,
        ctol: float | TensorType = 1e-7,
    ): ...

    @overload
    def __init__(
        self,
        lower: TensorType,
        upper: TensorType,
        constraints: Optional[Sequence[Constraint]] = None,
        ctol: float | TensorType = 1e-7,
    ): ...

    def __init__(
        self,
        lower: Sequence[float] | TensorType,
        upper: Sequence[float] | TensorType,
        constraints: Optional[Sequence[Constraint]] = None,
        ctol: float | TensorType = 1e-7,
    ):
        r"""
        If ``lower`` and ``upper`` are `Sequence`\ s of floats (such as lists or tuples),
        they will be converted to tensors of dtype `DEFAULT_DTYPE`.

        :param lower: The lower (inclusive) bounds of the box. Must have shape [D] for positive D,
            and if a tensor, must have float type.
        :param upper: The upper (inclusive) bounds of the box. Must have shape [D] for positive D,
            and if a tensor, must have float type.
        :param constraints: Sequence of explicit input constraints for this search space.
        :param ctol: Tolerance to use to check constraints satisfaction.
        :raise ValueError (or tf.errors.InvalidArgumentError): If any of the following are true:

            - ``lower`` and ``upper`` have invalid shapes.
            - ``lower`` and ``upper`` do not have the same floating point type.
            - ``upper`` is not greater or equal to ``lower`` across all dimensions.
        """

        tf.debugging.assert_shapes([(lower, ["D"]), (upper, ["D"])])
        tf.assert_rank(lower, 1)
        tf.assert_rank(upper, 1)
        tf.debugging.assert_non_negative(ctol, message="Tolerance must be non-negative")

        if isinstance(lower, Sequence):
            self._lower = tf.constant(lower, dtype=DEFAULT_DTYPE)
            self._upper = tf.constant(upper, dtype=DEFAULT_DTYPE)
        else:
            self._lower = tf.convert_to_tensor(lower)
            self._upper = tf.convert_to_tensor(upper)

            tf.debugging.assert_same_float_dtype([self._lower, self._upper])

        tf.debugging.assert_less_equal(self._lower, self._upper)

        self._dimension = tf.shape(self._upper)[-1]

        if constraints is None:
            self._constraints: Sequence[Constraint] = []
        else:
            self._constraints = constraints
        self._ctol = ctol

    def __repr__(self) -> str:
        """"""
        return f"Box({self._lower!r}, {self._upper!r}, {self._constraints!r}, {self._ctol!r})"

    @property
    def has_bounds(self) -> bool:
        return True

    @property
    def lower(self) -> tf.Tensor:
        """The lower bounds of the box."""
        return self._lower

    @property
    def upper(self) -> tf.Tensor:
        """The upper bounds of the box."""
        return self._upper

    @property
    def dimension(self) -> TensorType:
        """The number of inputs in this search space."""
        return self._dimension

    @property
    def constraints(self) -> Sequence[Constraint]:
        """The sequence of explicit constraints specified in this search space."""
        return self._constraints

    def _contains(self, value: TensorType) -> TensorType:
        """
        For each point in ``value``, return `True` if the point is a member of this search space,
        else `False`. A point is a member if all of its coordinates lie in the closed intervals
        bounded by the lower and upper bounds.

        :param value: A point or points to check for membership of this :class:`SearchSpace`.
        :return: A boolean array showing membership for each point in value.
        """
        return tf.reduce_all(value >= self._lower, axis=-1) & tf.reduce_all(
            value <= self._upper, axis=-1
        )

    def _sample(self, num_samples: int, seed: Optional[int] = None) -> TensorType:
        # Internal common method to sample randomly from the space.
        dim = tf.shape(self._lower)[-1]
        return tf.random.uniform(
            (num_samples, dim),
            minval=self._lower,
            maxval=self._upper,
            dtype=self._lower.dtype,
            seed=seed,
        )

    def sample(self, num_samples: int, seed: Optional[int] = None) -> TensorType:
        """
        Sample randomly from the space.

        :param num_samples: The number of points to sample from this search space.
        :param seed: Random seed for reproducibility.
        :return: ``num_samples`` i.i.d. random points, sampled uniformly,
            from this search space with shape '[num_samples, D]' , where D is the search space
            dimension.
        """
        tf.debugging.assert_non_negative(num_samples)
        if seed is not None:  # ensure reproducibility
            tf.random.set_seed(seed)
        return self._sample(num_samples, seed)

    def _sample_halton(
        self,
        start: int,
        num_samples: int,
        seed: Optional[int] = None,
    ) -> TensorType:
        # Internal common method to sample from the space using a Halton sequence.
        tf.debugging.assert_non_negative(num_samples)
        if num_samples == 0:
            return tf.constant([], dtype=self._lower.dtype)
        if seed is not None:  # ensure reproducibility
            tf.random.set_seed(seed)
        dim = tf.shape(self._lower)[-1]
        sequence_indices = tf.range(start=start, limit=start + num_samples, dtype=tf.int32)
        return (self._upper - self._lower) * tfp.mcmc.sample_halton_sequence(
            dim=dim, sequence_indices=sequence_indices, dtype=self._lower.dtype, seed=seed
        ) + self._lower

    def sample_halton(self, num_samples: int, seed: Optional[int] = None) -> TensorType:
        """
        Sample from the space using a Halton sequence. The resulting samples are guaranteed to be
        diverse and are reproducible by using the same choice of ``seed``.

        :param num_samples: The number of points to sample from this search space.
        :param seed: Random seed for the halton sequence
        :return: ``num_samples`` of points, using halton sequence with shape '[num_samples, D]' ,
            where D is the search space dimension.
        """
        return self._sample_halton(0, num_samples, seed)

    def sample_sobol(self, num_samples: int, skip: Optional[int] = None) -> TensorType:
        """
        Sample a diverse set from the space using a Sobol sequence.
        If ``skip`` is specified, then the resulting samples are reproducible.

        :param num_samples: The number of points to sample from this search space.
        :param skip: The number of initial points of the Sobol sequence to skip
        :return: ``num_samples`` of points, using sobol sequence with shape '[num_samples, D]' ,
            where D is the search space dimension.
        """
        tf.debugging.assert_non_negative(num_samples)
        if num_samples == 0:
            return tf.constant([], dtype=self._lower.dtype)
        if skip is None:  # generate random skip
            skip = tf.random.uniform([1], maxval=2**16, dtype=tf.int32)[0]
        dim = tf.shape(self._lower)[-1]
        return (self._upper - self._lower) * tf.math.sobol_sample(
            dim=dim, num_results=num_samples, dtype=self._lower.dtype, skip=skip
        ) + self._lower

    def _sample_feasible_loop(
        self,
        num_samples: int,
        sampler: Callable[[], TensorType],
        max_tries: int = 100,
    ) -> TensorType:
        """
        Rejection sampling using provided callable. Try ``max_tries`` number of times to find
        ``num_samples`` feasible points.

        :param num_samples: The number of feasible points to sample from this search space.
        :param sampler: Callable to return samples. Called potentially multiple times.
        :param max_tries: Maximum attempts to sample the requested number of points.
        :return: ``num_samples`` feasible points sampled using ``sampler``.
        :raise SampleTimeoutError: If ``max_tries`` are exhausted before ``num_samples`` are
            sampled.
        """
        xs = []
        count = 0
        tries = 0
        while count < num_samples and tries < max_tries:
            tries += 1
            xi = sampler()
            mask = self.is_feasible(xi)
            xo = tf.boolean_mask(xi, mask)
            xs.append(xo)
            count += xo.shape[0]

        if count < num_samples:
            raise SampleTimeoutError(
                f"""Failed to sample {num_samples} feasible point(s), even after {tries} attempts.
                    Sampled only {count} feasible point(s)."""
            )

        xs = tf.concat(xs, axis=0)[:num_samples]
        return xs

    def sample_feasible(
        self, num_samples: int, seed: Optional[int] = None, max_tries: int = 100
    ) -> TensorType:
        """
        Sample feasible points randomly from the space.

        :param num_samples: The number of feasible points to sample from this search space.
        :param seed: Random seed for reproducibility.
        :param max_tries: Maximum attempts to sample the requested number of points.
        :return: ``num_samples`` i.i.d. random points, sampled uniformly,
            from this search space with shape '[num_samples, D]' , where D is the search space
            dimension.
        :raise SampleTimeoutError: If ``max_tries`` are exhausted before ``num_samples`` are
            sampled.
        """
        tf.debugging.assert_non_negative(num_samples)

        # Without constraints or zero-num-samples use the normal sample method directly.
        if not self.has_constraints or num_samples == 0:
            return self.sample(num_samples, seed)

        if seed is not None:  # ensure reproducibility
            tf.random.set_seed(seed)

        def _sampler() -> TensorType:
            return self._sample(num_samples, seed)

        return self._sample_feasible_loop(num_samples, _sampler, max_tries)

    def sample_halton_feasible(
        self, num_samples: int, seed: Optional[int] = None, max_tries: int = 100
    ) -> TensorType:
        """
        Sample feasible points from the space using a Halton sequence. The resulting samples are
        guaranteed to be diverse and are reproducible by using the same choice of ``seed``.

        :param num_samples: The number of feasible points to sample from this search space.
        :param seed: Random seed for the halton sequence
        :param max_tries: Maximum attempts to sample the requested number of points.
        :return: ``num_samples`` of points, using halton sequence with shape '[num_samples, D]' ,
            where D is the search space dimension.
        :raise SampleTimeoutError: If ``max_tries`` are exhausted before ``num_samples`` are
            sampled.
        """
        tf.debugging.assert_non_negative(num_samples)

        # Without constraints or zero-num-samples use the normal sample method directly.
        if not self.has_constraints or num_samples == 0:
            return self.sample_halton(num_samples, seed)

        start = 0

        def _sampler() -> TensorType:
            nonlocal start
            # Global seed is set on every call in _sample_halton() so that we always sample from
            # the same (randomised) sequence, and skip the relevant number of beginning samples.
            samples = self._sample_halton(start, num_samples, seed)
            start += num_samples
            return samples

        return self._sample_feasible_loop(num_samples, _sampler, max_tries)

    def sample_sobol_feasible(
        self, num_samples: int, skip: Optional[int] = None, max_tries: int = 100
    ) -> TensorType:
        """
        Sample a diverse set of feasible points from the space using a Sobol sequence.
        If ``skip`` is specified, then the resulting samples are reproducible.

        :param num_samples: The number of feasible points to sample from this search space.
        :param skip: The number of initial points of the Sobol sequence to skip
        :param max_tries: Maximum attempts to sample the requested number of points.
        :return: ``num_samples`` of points, using sobol sequence with shape '[num_samples, D]' ,
            where D is the search space dimension.
        :raise SampleTimeoutError: If ``max_tries`` are exhausted before ``num_samples`` are
            sampled.
        """
        tf.debugging.assert_non_negative(num_samples)

        # Without constraints or zero-num-samples use the normal sample method directly.
        if not self.has_constraints or num_samples == 0:
            return self.sample_sobol(num_samples, skip)

        if skip is None:  # generate random skip
            skip = tf.random.uniform([1], maxval=2**16, dtype=tf.int32)[0]
        _skip: TensorType = skip  # To keep mypy happy.

        def _sampler() -> TensorType:
            nonlocal _skip
            samples = self.sample_sobol(num_samples, skip=_skip)
            # Skip the relevant number of beginning samples from previous iterations.
            _skip += num_samples
            return samples

        return self._sample_feasible_loop(num_samples, _sampler, max_tries)

    def product(self, other: Box) -> Box:
        r"""
        Return the Cartesian product of the two :class:`Box`\ es (concatenating their respective
        lower and upper bounds). For example:

            >>> unit_interval = Box([0.0], [1.0])
            >>> square_at_origin = Box([-2.0, -2.0], [2.0, 2.0])
            >>> new_box = unit_interval * square_at_origin
            >>> new_box.lower.numpy()
            array([ 0., -2., -2.])
            >>> new_box.upper.numpy()
            array([1., 2., 2.])

        :param other: A :class:`Box` with bounds of the same type as this :class:`Box`.
        :return: The Cartesian product of the two :class:`Box`\ es.
        :raise TypeError: If the bounds of one :class:`Box` have different dtypes to those of
            the other :class:`Box`.
        """
        if self.lower.dtype is not other.lower.dtype:
            return NotImplemented

        product_lower_bound = tf.concat([self._lower, other.lower], axis=-1)
        product_upper_bound = tf.concat([self._upper, other.upper], axis=-1)

        return Box(product_lower_bound, product_upper_bound)

    def __eq__(self, other: object) -> bool:
        """
        :param other: A search space.
        :return: Whether the search space is identical to this one.
        """
        if not isinstance(other, Box):
            return NotImplemented
        return bool(
            tf.reduce_all(self.lower == other.lower)
            and tf.reduce_all(self.upper == other.upper)
            # Constraints match only if they are exactly the same (in the same order).
            and self._constraints == other._constraints
        )

    def constraints_residuals(self, points: TensorType) -> TensorType:
        """
        Return residuals for all the constraints in this :class:`SearchSpace`.

        :param points: The points to get the residuals for, with shape [..., D].
        :return: A tensor of all the residuals with shape [..., C], where C is the total number of
            constraints.
        """
        residuals = [constraint.residual(points) for constraint in self._constraints]
        residuals = tf.concat(residuals, axis=-1)
        return residuals

    def is_feasible(self, points: TensorType) -> TensorType:
        """
        Checks if points satisfy the explicit constraints of this :class:`SearchSpace`.
        Note membership of the search space is not checked.

        :param points: The points to check constraints feasibility for, with shape [..., D].
        :return: A tensor of booleans. Returns `True` for each point if it is feasible in this
            search space, else `False`.
        """
        return tf.math.reduce_all(self.constraints_residuals(points) >= -self._ctol, axis=-1)

    @property
    def has_constraints(self) -> bool:
        """Returns `True` if this search space has any explicit constraints specified."""
        return len(self._constraints) > 0


class CollectionSearchSpace(SearchSpace):
    r"""
    An abstract :class:`SearchSpace` consisting of a collection of multiple :class:`SearchSpace`
    objects, each with a unique tag. This class provides functionality for accessing each individual
    space.

    Note that the individual spaces are not combined in any way.
    """

    def __init__(self, spaces: Sequence[SearchSpace], tags: Optional[Sequence[str]] = None):
        r"""
        Build a :class:`CollectionSearchSpace` from a list ``spaces`` of other spaces. If
        ``tags`` are provided then they form the identifiers of the subspaces, otherwise the
        subspaces are labelled numerically.

        :param spaces: A sequence of :class:`SearchSpace` objects representing the space's subspaces
        :param tags: An optional list of tags giving the unique identifiers of
            the space's subspaces.
        :raise ValueError (or tf.errors.InvalidArgumentError): If ``spaces`` has a different
            length to ``tags`` when ``tags`` is provided or if ``tags`` contains duplicates.
        """

        number_of_subspaces = len(spaces)
        if tags is None:
            tags = [str(index) for index in range(number_of_subspaces)]
        else:
            number_of_tags = len(tags)
            tf.debugging.assert_equal(
                number_of_tags,
                number_of_subspaces,
                message=f"""
                    Number of tags must match number of subspaces but
                    received {number_of_tags} tags and {number_of_subspaces} subspaces.
                """,
            )
            number_of_unique_tags = len(set(tags))
            tf.debugging.assert_equal(
                number_of_tags,
                number_of_unique_tags,
                message=f"Subspace names must be unique but received {tags}.",
            )

        self._spaces = dict(zip(tags, spaces))
        self._tags = tuple(tags)  # avoid accidental modification by users

    def __repr__(self) -> str:
        return f"""{self.__class__.__name__}(spaces =
                {[self.get_subspace(tag) for tag in self.subspace_tags]},
                tags = {self.subspace_tags})
                """

    @property
    def has_bounds(self) -> bool:
        """Whether the search space has meaningful numerical bounds."""
        return all(self.get_subspace(tag).has_bounds for tag in self.subspace_tags)

    @property
    def subspace_lower(self) -> Sequence[TensorType]:
        """The lowest values taken by each space dimension, in the same order as specified when
        initializing the space."""
        return [self.get_subspace(tag).lower for tag in self.subspace_tags]

    @property
    def subspace_upper(self) -> Sequence[TensorType]:
        """The highest values taken by each space dimension, in the same order as specified when
        initializing the space."""
        return [self.get_subspace(tag).upper for tag in self.subspace_tags]

    @property
    def subspace_tags(self) -> tuple[str, ...]:
        """Return the names of the subspaces contained in this space."""
        return self._tags

    @property
    def subspace_dimension(self) -> Sequence[TensorType]:
        """The number of inputs in each subspace, in the same order as specified when initializing
        the space."""
        return [self.get_subspace(tag).dimension for tag in self.subspace_tags]

    def get_subspace(self, tag: str) -> SearchSpace:
        """
        Return the domain of a particular subspace.

        :param tag: The tag specifying the target subspace.
        :return: Target subspace.
        """
        tf.debugging.assert_equal(
            tag in self.subspace_tags,
            True,
            message=f"""
                Attempted to access a subspace that does not exist. This space only contains
                subspaces with the tags {self.subspace_tags} but received {tag}.
                """,
        )
        return self._spaces[tag]

    def subspace_sample(self, num_samples: int, seed: Optional[int] = None) -> Sequence[TensorType]:
        """
        Sample randomly from the space by sampling from each subspace
        and returning the resulting samples in the same order as specified when initializing
        the space.

        :param num_samples: The number of points to sample from each subspace.
        :param seed: Optional tf.random seed.
        :return: ``num_samples`` i.i.d. random points, sampled uniformly,
            from each search subspace with shape '[num_samples, D]' , where D is the search space
            dimension.
        """
        tf.debugging.assert_non_negative(num_samples)
        if seed is not None:  # ensure reproducibility
            tf.random.set_seed(seed)
        return [
            # ensure subspaces (which may be identical) don't all use the same seed
            self._spaces[tag].sample(num_samples, seed=None if seed is None else seed + i)
            for i, tag in enumerate(self._tags)
        ]

    def __eq__(self, other: object) -> bool:
        """
        :param other: A search space.
        :return: Whether the search space is identical to this one.
        """
        if not isinstance(other, type(self)):
            return NotImplemented
        return self._tags == other._tags and self._spaces == other._spaces


class TaggedProductSearchSpace(CollectionSearchSpace, HasOneHotEncoder):
    r"""
    Product :class:`SearchSpace` consisting of a product of
    multiple :class:`SearchSpace`. This class provides functionality for
    accessing either the resulting combined search space or each individual space.
    This class is useful for defining mixed search spaces, for example:

        context_space = DiscreteSearchSpace(tf.constant([[-0.5, 0.5]]))
        decision_space = Box([-1, -2], [2, 3])
        mixed_space = TaggedProductSearchSpace(spaces=[context_space, decision_space])

    Note: the dtype of all the component search spaces must be the same.

    Note that this class assumes that individual points in product spaces are
    represented with their inputs in the same order as specified when initializing
    the space.
    """

    def __init__(self, spaces: Sequence[SearchSpace], tags: Optional[Sequence[str]] = None):
        r"""
        Build a :class:`TaggedProductSearchSpace` from a list ``spaces`` of other spaces. If
        ``tags`` are provided then they form the identifiers of the subspaces, otherwise the
        subspaces are labelled numerically.

        :param spaces: A sequence of :class:`SearchSpace` objects representing the space's subspaces
        :param tags: An optional list of tags giving the unique identifiers of
            the space's subspaces.
        :raise ValueError (or tf.errors.InvalidArgumentError): If ``spaces`` has a different
            length to ``tags`` when ``tags`` is provided or if ``tags`` contains duplicates.
        """

        super().__init__(spaces, tags)
        subspace_sizes = self.subspace_dimension

        self._subspace_sizes_by_tag = dict(zip(self._tags, subspace_sizes))

        self._subspace_starting_indices = dict(
            zip(self._tags, tf.cumsum(subspace_sizes, exclusive=True))
        )

        self._dimension = tf.cast(tf.reduce_sum(subspace_sizes), dtype=tf.int32)

    @property
    @check_shapes("return: [D]")
    def lower(self) -> TensorType:
        """The lowest values taken by each space dimension, concatenated across subspaces."""
        lower_for_each_subspace = self.subspace_lower
        return (
            tf.concat(lower_for_each_subspace, axis=-1)
            if lower_for_each_subspace
            else tf.constant([], dtype=DEFAULT_DTYPE)
        )

    @property
    @check_shapes("return: [D]")
    def upper(self) -> TensorType:
        """The highest values taken by each space dimension, concatenated across subspaces."""
        upper_for_each_subspace = self.subspace_upper
        return (
            tf.concat(upper_for_each_subspace, axis=-1)
            if upper_for_each_subspace
            else tf.constant([], dtype=DEFAULT_DTYPE)
        )

    @property
    @check_shapes("return: []")
    def dimension(self) -> TensorType:
        """The number of inputs in this product search space."""
        return self._dimension

    def fix_subspace(self, tag: str, values: TensorType) -> TaggedProductSearchSpace:
        """
        Return a new :class:`TaggedProductSearchSpace` with the specified subspace replaced with
        a :class:`DiscreteSearchSpace` containing ``values`` as its points. This is useful if you
        wish to restrict subspaces to sets of representative points.

        :param tag: The tag specifying the target subspace.
        :param values: The  values used to populate the new discrete subspace.z
        :return: New :class:`TaggedProductSearchSpace` with the specified subspace replaced with
            a :class:`DiscreteSearchSpace` containing ``values`` as its points.
        """

        new_spaces = [
            self.get_subspace(t) if t != tag else DiscreteSearchSpace(points=values)
            for t in self.subspace_tags
        ]

        return TaggedProductSearchSpace(spaces=new_spaces, tags=self.subspace_tags)

    def get_subspace_component(self, tag: str, values: TensorType) -> TensorType:
        """
        Returns the components of ``values`` lying in a particular subspace.

        :param tag: Subspace tag.
        :param values: Points from the :class:`TaggedProductSearchSpace` of shape [N,Dprod].
        :return: The sub-components of ``values`` lying in the specified subspace, of shape
            [N, Dsub], where Dsub is the dimensionality of the specified subspace.
        """

        starting_index_of_subspace = self._subspace_starting_indices[tag]
        ending_index_of_subspace = starting_index_of_subspace + self._subspace_sizes_by_tag[tag]
        return values[..., starting_index_of_subspace:ending_index_of_subspace]

    def _contains(self, value: TensorType) -> TensorType:
        """
        Return `True` if ``value`` is a member of this search space, else `False`. A point is a
        member if each of its subspace components lie in each subspace.

        Recall that individual points in product spaces are represented with their inputs in the
        same order as specified when initializing the space.

        :param value: A point to check for membership of this :class:`SearchSpace`.
        :return: `True` if ``value`` is a member of this search space, else `False`. May return a
            scalar boolean `TensorType` instead of the `bool` itself.
        :raise ValueError (or tf.errors.InvalidArgumentError): If ``value`` has a different
            dimensionality from the search space.
        """
        in_each_subspace = [
            self._spaces[tag].contains(self.get_subspace_component(tag, value))
            for tag in self._tags
        ]
        return tf.reduce_all(in_each_subspace, axis=0)

    @check_shapes("return: [num_samples, D]")
    def sample(self, num_samples: int, seed: Optional[int] = None) -> TensorType:
        """
        Sample randomly from the space by sampling from each subspace
        and concatenating the resulting samples.

        :param num_samples: The number of points to sample from this search space.
        :param seed: Optional tf.random seed.
        :return: ``num_samples`` i.i.d. random points, sampled uniformly,
            from this search space with shape '[num_samples, D]' , where D is the search space
            dimension.
        """
        subspace_samples = self.subspace_sample(num_samples, seed)
        return tf.concat(subspace_samples, -1)

    def product(self, other: TaggedProductSearchSpace) -> TaggedProductSearchSpace:
        r"""
        Return the Cartesian product of the two :class:`TaggedProductSearchSpace`\ s,
        building a tree of :class:`TaggedProductSearchSpace`\ s.

        :param other: A search space of the same type as this search space.
        :return: The Cartesian product of this search space with the ``other``.
        """
        return TaggedProductSearchSpace(spaces=[self, other])

    @property
    def one_hot_encoder(self) -> EncoderFunction:
        """An encoder that one-hot-encodes all subpsaces that support it (and leaves
        the other subspaces unchanged)."""

        def encoder(x: TensorType) -> TensorType:
            components = []
            for tag in self.subspace_tags:
                component = self.get_subspace_component(tag, x)
                space = self.get_subspace(tag)
                if isinstance(space, HasOneHotEncoder):
                    component = space.one_hot_encoder(component)
                components.append(component)
            return tf.concat(components, axis=-1)

        return encoder


class TaggedMultiSearchSpace(CollectionSearchSpace):
    r"""
    A :class:`SearchSpace` made up of a collection of multiple :class:`SearchSpace` subspaces,
    each with a unique tag. All subspaces must have the same dimensionality.

    Each subspace is treated as an independent space and not combined in any way. This class
    provides functionality for accessing all the subspaces at once by using the usual search space
    methods, as well as for accessing individual subspaces.

    When accessing all subspaces at once from this class (e.g. `lower()`, `upper()`, `sample()`),
    the returned tensors have an extra dimension corresponding to the subspaces.

    This class can be useful to represent a collection of search spaces that do not interact with
    each other. For example, it is used to implement batch trust region rules in the
    :class:`BatchTrustRegion` class.
    """

    def __init__(self, spaces: Sequence[SearchSpace], tags: Optional[Sequence[str]] = None):
        r"""
        Build a :class:`TaggedMultiSearchSpace` from a list ``spaces`` of other spaces. If
        ``tags`` are provided then they form the identifiers of the subspaces, otherwise the
        subspaces are labelled numerically.

        :param spaces: A sequence of :class:`SearchSpace` objects representing the space's subspaces
        :param tags: An optional list of tags giving the unique identifiers of
            the space's subspaces.
        :raise ValueError (or tf.errors.InvalidArgumentError): If ``spaces`` has a different
            length to ``tags`` when ``tags`` is provided or if ``tags`` contains duplicates.
        :raise ValueError (or tf.errors.InvalidArgumentError): If ``spaces`` has a different
            dimension to each other.
        """

        # At least one subspace is required.
        tf.debugging.assert_greater(
            len(spaces),
            0,
            message=f"""
                At least one subspace is required but received {len(spaces)}.
                """,
        )

        tf.debugging.assert_equal(
            len({int(space.dimension) for space in spaces}),
            1,
            message=f"""
                All subspaces must have the same dimension but received
                {[int(space.dimension) for space in spaces]}.
                """,
        )

        super().__init__(spaces, tags)

    @property
    @check_shapes("return: [V, D]")
    def lower(self) -> TensorType:
        """Returns the stacked lower bounds of all the subspaces.

        :return: The lower bounds of shape [V, D], where V is the number of subspaces and D is
            the dimensionality of each subspace.
        """
        return tf.stack(self.subspace_lower, axis=0)

    @property
    @check_shapes("return: [V, D]")
    def upper(self) -> TensorType:
        """Returns the stacked upper bounds of all the subspaces.

        :return: The upper bounds of shape [V, D], where V is the number of subspaces and D is
            the dimensionality of each subspace.
        """
        return tf.stack(self.subspace_upper, axis=0)

    @property
    @check_shapes("return: []")
    def dimension(self) -> TensorType:
        """The number of inputs in this search space."""
        return self.get_subspace(self.subspace_tags[0]).dimension

    @check_shapes("return: [num_samples, V, D]")
    def sample(self, num_samples: int, seed: Optional[int] = None) -> TensorType:
        """
        Sample randomly from the space by sampling from each subspace
        and returning the resulting samples stacked along the second axis in the same order as
        specified when initializing the space.

        :param num_samples: The number of points to sample from each subspace.
        :param seed: Optional tf.random seed.
        :return: ``num_samples`` i.i.d. random points, sampled uniformly,
            from each search subspace with shape '[num_samples, V, D]' , where V is the number of
            subspaces and D is the search space dimension.
        """
        return tf.stack(self.subspace_sample(num_samples, seed), axis=1)

    def _contains(self, value: TensorType) -> TensorType:
        """
        Return `True` if ``value`` is a member of this search space, else `False`. A point
        is a member if it is a member of any of the subspaces.

        :param value: A point or points to check for membership of this :class:`SearchSpace`.
        :return: A boolean array showing membership for each point in value.
        """
        return tf.reduce_any(
            [self.get_subspace(tag)._contains(value) for tag in self.subspace_tags], axis=0
        )

    def product(self, other: TaggedMultiSearchSpace) -> TaggedMultiSearchSpace:
        r"""
        Return a bigger collection of two :class:`TaggedMultiSearchSpace`\ s, regenerating the
        tags.

        :param other: A search space of the same type as this search space.
        :return: The product of this search space with the ``other``.
        """
        return TaggedMultiSearchSpace(
            spaces=tuple(self._spaces.values()) + tuple(other._spaces.values())
        )

    def discretize(self, num_samples: int) -> DiscreteSearchSpace:
        """
        :param num_samples: The number of points in the :class:`DiscreteSearchSpace`.
        :return: A discrete search space consisting of ``num_samples`` points sampled uniformly from
            this search space.
        :raise NotImplementedError: If this :class:`SearchSpace` has constraints.
        """
        if self.has_constraints:  # Constraints are not supported.
            raise NotImplementedError(
                "Discretization is currently not supported in the presence of constraints."
            )
        samples = self.sample(num_samples)  # Sample num_samples points from each subspace.
        samples = tf.reshape(samples, [-1, self.dimension])  # Flatten the samples across subspaces.
        samples = tf.random.shuffle(samples)[:num_samples]  # Randomly pick num_samples points.
        return DiscreteSearchSpace(points=samples)



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the :class:`BayesianOptimizer` class, used to perform Bayesian optimization.
"""

from __future__ import annotations

import copy
import traceback
from collections import Counter
from dataclasses import dataclass
from pathlib import Path
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Mapping,
    MutableMapping,
    Optional,
    TypeVar,
    cast,
    overload,
)

import absl
import dill
import numpy as np
import tensorflow as tf
from scipy.spatial.distance import pdist

from .acquisition.multi_objective import non_dominated
from .models.utils import optimize_model_and_save_result

try:
    import pandas as pd
    import seaborn as sns
except ModuleNotFoundError:
    pd = None
    sns = None

from . import logging
from .acquisition.rule import (
    AcquisitionRule,
    EfficientGlobalOptimization,
    LocalDatasetsAcquisitionRule,
)
from .acquisition.utils import with_local_datasets
from .data import Dataset
from .models import (
    ProbabilisticModel,
    SupportsCovarianceWithTopFidelity,
    TrainableProbabilisticModel,
)
from .objectives.utils import mk_batch_observer
from .observer import OBJECTIVE, Observer
from .space import SearchSpace
from .types import State, Tag, TensorType
from .utils import Err, Ok, Result, Timer
from .utils.misc import LocalizedTag, get_value_for_tag, ignoring_local_tags

StateType = TypeVar("StateType")
""" Unbound type variable. """

SearchSpaceType = TypeVar("SearchSpaceType", bound=SearchSpace)
""" Type variable bound to :class:`SearchSpace`. """

ProbabilisticModelType = TypeVar(
    "ProbabilisticModelType",
    bound=ProbabilisticModel,
    covariant=True,
)
""" Covariant type variable bound to :class:`ProbabilisticModel`. """

TrainableProbabilisticModelType = TypeVar(
    "TrainableProbabilisticModelType", bound=TrainableProbabilisticModel, contravariant=True
)
""" Contravariant type variable bound to :class:`TrainableProbabilisticModel`. """

EarlyStopCallback = Callable[
    [Mapping[Tag, Dataset], Mapping[Tag, TrainableProbabilisticModelType], Optional[StateType]],
    bool,
]
""" Early stop callback type, generic in the model and state types. """


@dataclass(frozen=True)
class Record(Generic[StateType, ProbabilisticModelType]):
    """Container to record the state of each step of the optimization process."""

    datasets: Mapping[Tag, Dataset]
    """ The known data from the observer. """

    models: Mapping[Tag, ProbabilisticModelType]
    """ The models over the :attr:`datasets`. """

    acquisition_state: StateType | None
    """ The acquisition state. """

    @property
    def dataset(self) -> Dataset:
        """The dataset when there is just one dataset."""
        # Ignore local datasets.
        datasets: Mapping[Tag, Dataset] = ignoring_local_tags(self.datasets)
        if len(datasets) == 1:
            return next(iter(datasets.values()))
        else:
            raise ValueError(f"Expected a single dataset, found {len(datasets)}")

    @property
    def model(self) -> ProbabilisticModelType:
        """The model when there is just one dataset."""
        # Ignore local models.
        models: Mapping[Tag, ProbabilisticModelType] = ignoring_local_tags(self.models)
        if len(models) == 1:
            return next(iter(models.values()))
        else:
            raise ValueError(f"Expected a single model, found {len(models)}")

    def save(self, path: Path | str) -> FrozenRecord[StateType, ProbabilisticModelType]:
        """Save the record to disk. Will overwrite any existing file at the same path."""
        Path(path).parent.mkdir(exist_ok=True, parents=True)
        with open(path, "wb") as f:
            dill.dump(self, f, dill.HIGHEST_PROTOCOL)
        return FrozenRecord(Path(path))


@dataclass(frozen=True)
class FrozenRecord(Generic[StateType, ProbabilisticModelType]):
    """
    A Record container saved on disk.

    Note that records are saved via pickling and are therefore neither portable nor secure.
    Only open frozen records generated on the same system.
    """

    path: Path
    """ The path to the pickled Record. """

    def load(self) -> Record[StateType, ProbabilisticModelType]:
        """Load the record into memory."""
        with open(self.path, "rb") as f:
            return dill.load(f)

    @property
    def datasets(self) -> Mapping[Tag, Dataset]:
        """The known data from the observer."""
        return self.load().datasets

    @property
    def models(self) -> Mapping[Tag, ProbabilisticModelType]:
        """The models over the :attr:`datasets`."""
        return self.load().models

    @property
    def acquisition_state(self) -> StateType | None:
        """The acquisition state."""
        return self.load().acquisition_state

    @property
    def dataset(self) -> Dataset:
        """The dataset when there is just one dataset."""
        return self.load().dataset

    @property
    def model(self) -> ProbabilisticModelType:
        """The model when there is just one dataset."""
        return self.load().model


# this should be a generic NamedTuple, but mypy doesn't support them
#  https://github.com/python/mypy/issues/685
@dataclass(frozen=True)
class OptimizationResult(Generic[StateType, ProbabilisticModelType]):
    """The final result, and the historical data of the optimization process."""

    final_result: Result[Record[StateType, ProbabilisticModelType]]
    """
    The final result of the optimization process. This contains either a :class:`Record` or an
    exception.
    """

    history: list[
        Record[StateType, ProbabilisticModelType] | FrozenRecord[StateType, ProbabilisticModelType]
    ]
    r"""
    The history of the :class:`Record`\ s from each step of the optimization process. These
    :class:`Record`\ s are created at the *start* of each loop, and as such will never
    include the :attr:`final_result`. The records may be either in memory or on disk.
    """

    @staticmethod
    def step_filename(step: int, num_steps: int) -> str:
        """Default filename for saved optimization steps."""
        return f"step.{step:0{len(str(num_steps - 1))}d}.pickle"

    STEP_GLOB: ClassVar[str] = "step.*.pickle"
    RESULTS_FILENAME: ClassVar[str] = "results.pickle"

    def astuple(
        self,
    ) -> tuple[
        Result[Record[StateType, ProbabilisticModelType]],
        list[
            Record[StateType, ProbabilisticModelType]
            | FrozenRecord[StateType, ProbabilisticModelType]
        ],
    ]:
        """
        **Note:** In contrast to the standard library function :func:`dataclasses.astuple`, this
        method does *not* deepcopy instance attributes.

        :return: The :attr:`final_result` and :attr:`history` as a 2-tuple.
        """
        return self.final_result, self.history

    @property
    def is_ok(self) -> bool:
        """`True` if the final result contains a :class:`Record`."""
        return self.final_result.is_ok

    @property
    def is_err(self) -> bool:
        """`True` if the final result contains an exception."""
        return self.final_result.is_err

    def try_get_final_datasets(self) -> Mapping[Tag, Dataset]:
        """
        Convenience method to attempt to get the final data.

        :return: The final data, if the optimization completed successfully.
        :raise Exception: If an exception occurred during optimization.
        """
        return self.final_result.unwrap().datasets

    def try_get_final_dataset(self) -> Dataset:
        """
        Convenience method to attempt to get the final data for a single dataset run.

        :return: The final data, if the optimization completed successfully.
        :raise Exception: If an exception occurred during optimization.
        :raise ValueError: If the optimization was not a single dataset run.
        """
        datasets = self.try_get_final_datasets()
        # Ignore local datasets.
        datasets = ignoring_local_tags(datasets)
        if len(datasets) == 1:
            return next(iter(datasets.values()))
        else:
            raise ValueError(f"Expected a single dataset, found {len(datasets)}")

    def try_get_optimal_point(self) -> tuple[TensorType, TensorType, TensorType]:
        """
        Convenience method to attempt to get the optimal point for a single dataset,
        single objective run.

        :return: Tuple of the optimal query point, observation and its index.
        """
        dataset = self.try_get_final_dataset()
        if tf.rank(dataset.observations) != 2 or dataset.observations.shape[1] != 1:
            raise ValueError("Expected a single objective")
        if tf.reduce_any(
            [
                isinstance(model, SupportsCovarianceWithTopFidelity)
                for model in self.try_get_final_models()
            ]
        ):
            raise ValueError("Expected single fidelity models")
        arg_min_idx = tf.squeeze(tf.argmin(dataset.observations, axis=0))
        return dataset.query_points[arg_min_idx], dataset.observations[arg_min_idx], arg_min_idx

    def try_get_final_models(self) -> Mapping[Tag, ProbabilisticModelType]:
        """
        Convenience method to attempt to get the final models.

        :return: The final models, if the optimization completed successfully.
        :raise Exception: If an exception occurred during optimization.
        """
        return self.final_result.unwrap().models

    def try_get_final_model(self) -> ProbabilisticModelType:
        """
        Convenience method to attempt to get the final model for a single model run.

        :return: The final model, if the optimization completed successfully.
        :raise Exception: If an exception occurred during optimization.
        :raise ValueError: If the optimization was not a single model run.
        """
        models = self.try_get_final_models()
        # Ignore local models.
        models = ignoring_local_tags(models)
        if len(models) == 1:
            return next(iter(models.values()))
        else:
            raise ValueError(f"Expected single model, found {len(models)}")

    @property
    def loaded_history(self) -> list[Record[StateType, ProbabilisticModelType]]:
        """The history of the optimization process loaded into memory."""
        return [record if isinstance(record, Record) else record.load() for record in self.history]

    def save_result(self, path: Path | str) -> None:
        """Save the final result to disk. Will overwrite any existing file at the same path."""
        Path(path).parent.mkdir(exist_ok=True, parents=True)
        with open(path, "wb") as f:
            dill.dump(self.final_result, f, dill.HIGHEST_PROTOCOL)

    def save(self, base_path: Path | str) -> None:
        """Save the optimization result to disk. Will overwrite existing files at the same path."""
        path = Path(base_path)
        num_steps = len(self.history)
        self.save_result(path / self.RESULTS_FILENAME)
        for i, record in enumerate(self.loaded_history):
            record_path = path / self.step_filename(i, num_steps)
            record.save(record_path)

    @classmethod
    def from_path(
        cls, base_path: Path | str
    ) -> OptimizationResult[StateType, ProbabilisticModelType]:
        """Load a previously saved OptimizationResult."""
        try:
            with open(Path(base_path) / cls.RESULTS_FILENAME, "rb") as f:
                result = dill.load(f)
        except FileNotFoundError as e:
            result = Err(e)

        history: list[
            Record[StateType, ProbabilisticModelType]
            | FrozenRecord[StateType, ProbabilisticModelType]
        ] = [FrozenRecord(file) for file in sorted(Path(base_path).glob(cls.STEP_GLOB))]
        return cls(result, history)


class BayesianOptimizer(Generic[SearchSpaceType]):
    """
    This class performs Bayesian optimization, the data-efficient optimization of an expensive
    black-box *objective function* over some *search space*. Since we may not have access to the
    objective function itself, we speak instead of an *observer* that observes it.
    """

    def __init__(self, observer: Observer, search_space: SearchSpaceType):
        """
        :param observer: The observer of the objective function.
        :param search_space: The space over which to search. Must be a
            :class:`~trieste.space.SearchSpace`.
        """
        self._observer = observer
        self._search_space = search_space

    def __repr__(self) -> str:
        """"""
        return f"BayesianOptimizer({self._observer!r}, {self._search_space!r})"

    @overload
    def optimize(
        self,
        num_steps: int,
        datasets: Mapping[Tag, Dataset],
        models: Mapping[Tag, TrainableProbabilisticModel],
        *,
        track_state: bool = True,
        track_path: Optional[Path | str] = None,
        fit_model: bool = True,
        fit_initial_model: bool = True,
        early_stop_callback: Optional[
            EarlyStopCallback[TrainableProbabilisticModel, object]
        ] = None,
        start_step: int = 0,
    ) -> OptimizationResult[None, TrainableProbabilisticModel]: ...

    @overload
    def optimize(
        self,
        num_steps: int,
        datasets: Mapping[Tag, Dataset],
        models: Mapping[Tag, TrainableProbabilisticModelType],
        acquisition_rule: AcquisitionRule[
            TensorType, SearchSpaceType, TrainableProbabilisticModelType
        ],
        *,
        track_state: bool = True,
        track_path: Optional[Path | str] = None,
        fit_model: bool = True,
        fit_initial_model: bool = True,
        early_stop_callback: Optional[
            EarlyStopCallback[TrainableProbabilisticModelType, object]
        ] = None,
        start_step: int = 0,
        # this should really be OptimizationResult[None], but tf.Tensor is untyped so the type
        # checker can't differentiate between TensorType and State[S | None, TensorType], and
        # the return types clash. object is close enough to None that object will do.
    ) -> OptimizationResult[object, TrainableProbabilisticModelType]: ...

    @overload
    def optimize(
        self,
        num_steps: int,
        datasets: Mapping[Tag, Dataset],
        models: Mapping[Tag, TrainableProbabilisticModelType],
        acquisition_rule: AcquisitionRule[
            TensorType, SearchSpaceType, TrainableProbabilisticModelType
        ],
        *,
        track_state: bool = True,
        track_path: Optional[Path | str] = None,
        fit_model: bool = True,
        fit_initial_model: bool = True,
        early_stop_callback: Optional[
            EarlyStopCallback[TrainableProbabilisticModelType, object]
        ] = None,
        start_step: int = 0,
    ) -> OptimizationResult[object, TrainableProbabilisticModelType]: ...

    @overload
    def optimize(
        self,
        num_steps: int,
        datasets: Mapping[Tag, Dataset],
        models: Mapping[Tag, TrainableProbabilisticModelType],
        acquisition_rule: AcquisitionRule[
            State[StateType | None, TensorType], SearchSpaceType, TrainableProbabilisticModelType
        ],
        acquisition_state: StateType | None = None,
        *,
        track_state: bool = True,
        track_path: Optional[Path | str] = None,
        fit_model: bool = True,
        fit_initial_model: bool = True,
        early_stop_callback: Optional[
            EarlyStopCallback[TrainableProbabilisticModelType, StateType]
        ] = None,
        start_step: int = 0,
    ) -> OptimizationResult[StateType, TrainableProbabilisticModelType]: ...

    @overload
    def optimize(
        self,
        num_steps: int,
        datasets: Mapping[Tag, Dataset],
        models: Mapping[Tag, TrainableProbabilisticModelType],
        acquisition_rule: AcquisitionRule[
            State[StateType | None, TensorType], SearchSpaceType, TrainableProbabilisticModelType
        ],
        acquisition_state: StateType | None = None,
        *,
        track_state: bool = True,
        track_path: Optional[Path | str] = None,
        fit_model: bool = True,
        fit_initial_model: bool = True,
        early_stop_callback: Optional[
            EarlyStopCallback[TrainableProbabilisticModelType, StateType]
        ] = None,
        start_step: int = 0,
    ) -> OptimizationResult[StateType, TrainableProbabilisticModelType]: ...

    @overload
    def optimize(
        self,
        num_steps: int,
        datasets: Dataset,
        models: TrainableProbabilisticModel,
        *,
        track_state: bool = True,
        track_path: Optional[Path | str] = None,
        fit_model: bool = True,
        fit_initial_model: bool = True,
        early_stop_callback: Optional[
            EarlyStopCallback[TrainableProbabilisticModel, object]
        ] = None,
        start_step: int = 0,
    ) -> OptimizationResult[None, TrainableProbabilisticModel]: ...

    @overload
    def optimize(
        self,
        num_steps: int,
        datasets: Dataset,
        models: TrainableProbabilisticModelType,
        acquisition_rule: AcquisitionRule[
            TensorType, SearchSpaceType, TrainableProbabilisticModelType
        ],
        *,
        track_state: bool = True,
        track_path: Optional[Path | str] = None,
        fit_model: bool = True,
        fit_initial_model: bool = True,
        early_stop_callback: Optional[
            EarlyStopCallback[TrainableProbabilisticModelType, object]
        ] = None,
        start_step: int = 0,
    ) -> OptimizationResult[object, TrainableProbabilisticModelType]: ...

    @overload
    def optimize(
        self,
        num_steps: int,
        datasets: Dataset,
        models: TrainableProbabilisticModelType,
        acquisition_rule: AcquisitionRule[
            TensorType, SearchSpaceType, TrainableProbabilisticModelType
        ],
        *,
        track_state: bool = True,
        track_path: Optional[Path | str] = None,
        fit_model: bool = True,
        fit_initial_model: bool = True,
        early_stop_callback: Optional[
            EarlyStopCallback[TrainableProbabilisticModelType, object]
        ] = None,
        start_step: int = 0,
    ) -> OptimizationResult[object, TrainableProbabilisticModelType]: ...

    @overload
    def optimize(
        self,
        num_steps: int,
        datasets: Dataset,
        models: TrainableProbabilisticModelType,
        acquisition_rule: AcquisitionRule[
            State[StateType | None, TensorType], SearchSpaceType, TrainableProbabilisticModelType
        ],
        acquisition_state: StateType | None = None,
        *,
        track_state: bool = True,
        track_path: Optional[Path | str] = None,
        fit_model: bool = True,
        fit_initial_model: bool = True,
        early_stop_callback: Optional[
            EarlyStopCallback[TrainableProbabilisticModelType, StateType]
        ] = None,
        start_step: int = 0,
    ) -> OptimizationResult[StateType, TrainableProbabilisticModelType]: ...

    @overload
    def optimize(
        self,
        num_steps: int,
        datasets: Dataset,
        models: TrainableProbabilisticModelType,
        acquisition_rule: AcquisitionRule[
            State[StateType | None, TensorType], SearchSpaceType, TrainableProbabilisticModelType
        ],
        acquisition_state: StateType | None = None,
        *,
        track_state: bool = True,
        track_path: Optional[Path | str] = None,
        fit_model: bool = True,
        fit_initial_model: bool = True,
        early_stop_callback: Optional[
            EarlyStopCallback[TrainableProbabilisticModelType, StateType]
        ] = None,
        start_step: int = 0,
    ) -> OptimizationResult[StateType, TrainableProbabilisticModelType]: ...

    def optimize(
        self,
        num_steps: int,
        datasets: Mapping[Tag, Dataset] | Dataset,
        models: Mapping[Tag, TrainableProbabilisticModelType] | TrainableProbabilisticModelType,
        acquisition_rule: (
            AcquisitionRule[
                TensorType | State[StateType | None, TensorType],
                SearchSpaceType,
                TrainableProbabilisticModelType,
            ]
            | None
        ) = None,
        acquisition_state: StateType | None = None,
        *,
        track_state: bool = True,
        track_path: Optional[Path | str] = None,
        fit_model: bool = True,
        fit_initial_model: bool = True,
        early_stop_callback: Optional[
            EarlyStopCallback[TrainableProbabilisticModelType, StateType]
        ] = None,
        start_step: int = 0,
    ) -> (
        OptimizationResult[StateType, TrainableProbabilisticModelType]
        | OptimizationResult[None, TrainableProbabilisticModelType]
    ):
        """
        Attempt to find the minimizer of the ``observer`` in the ``search_space`` (both specified at
        :meth:`__init__`). This is the central implementation of the Bayesian optimization loop.

        For each step in ``num_steps``, this method:
            - Finds the next points with which to query the ``observer`` using the
              ``acquisition_rule``'s :meth:`acquire` method, passing it the ``search_space``,
              ``datasets``, ``models``, and current acquisition state.
            - Queries the ``observer`` *once* at those points.
            - Updates the datasets and models with the data from the ``observer``.

        If any errors are raised during the optimization loop, this method will catch and return
        them instead and print a message (using `absl` at level `absl.logging.ERROR`).
        If ``track_state`` is enabled, then in addition to the final result, the history of the
        optimization process will also be returned. If ``track_path`` is also set, then
        the history and final result will be saved to disk rather than all being kept in memory.

        **Type hints:**
            - The ``acquisition_rule`` must use the same type of
              :class:`~trieste.space.SearchSpace` as specified in :meth:`__init__`.
            - The ``acquisition_state`` must be of the type expected by the ``acquisition_rule``.
              Any acquisition state in the optimization result will also be of this type.

        :param num_steps: The number of optimization steps to run.
        :param datasets: The known observer query points and observations for each tag.
        :param models: The model to use for each :class:`~trieste.data.Dataset` in
            ``datasets``.
        :param acquisition_rule: The acquisition rule, which defines how to search for a new point
            on each optimization step. Defaults to
            :class:`~trieste.acquisition.rule.EfficientGlobalOptimization` with default
            arguments. Note that if the default is used, this implies the tags must be
            `OBJECTIVE`, the search space can be any :class:`~trieste.space.SearchSpace`, and the
            acquisition state returned in the :class:`OptimizationResult` will be `None`.
        :param acquisition_state: The acquisition state to use on the first optimization step.
            This argument allows the caller to restore the optimization process from an existing
            :class:`Record`.
        :param track_state: If `True`, this method saves the optimization state at the start of each
            step. Models and acquisition state are copied using `copy.deepcopy`.
        :param track_path: If set, the optimization state is saved to disk at this path,
            rather than being copied in memory.
        :param fit_model: If `False` then we never fit the model during BO (e.g. if we
            are using a rule that doesn't rely on the models and don't want to waste computation).
        :param fit_initial_model: If `False` then we assume that the initial models have
            already been optimized on the datasets and so do not require optimization before
            the first optimization step.
        :param early_stop_callback: An optional callback that is evaluated with the current
            datasets, models and optimization state before every optimization step. If this
            returns `True` then the optimization loop is terminated early.
        :param start_step: The step number to start with. This number is removed from ``num_steps``
            and is useful for restarting previous computations.
        :return: An :class:`OptimizationResult`. The :attr:`final_result` element contains either
            the final optimization data, models and acquisition state, or, if an exception was
            raised while executing the optimization loop, it contains the exception raised. In
            either case, the :attr:`history` element is the history of the data, models and
            acquisition state at the *start* of each optimization step (up to and including any step
            that fails to complete). The history will never include the final optimization result.
        :raise ValueError: If any of the following are true:

            - ``num_steps`` is negative.
            - the keys in ``datasets`` and ``models`` do not match
            - ``datasets`` or ``models`` are empty
            - the default `acquisition_rule` is used and the tags are not `OBJECTIVE`.
        """
        # Copy the dataset so we don't change the one provided by the user.
        datasets = copy.deepcopy(datasets)

        if isinstance(datasets, Dataset):
            datasets = {OBJECTIVE: datasets}
        if not isinstance(models, Mapping):
            models = {OBJECTIVE: models}

        filtered_datasets = datasets
        # reassure the type checker that everything is tagged
        datasets = cast(Dict[Tag, Dataset], datasets)
        models = cast(Dict[Tag, TrainableProbabilisticModelType], models)

        if num_steps < 0:
            raise ValueError(f"num_steps must be at least 0, got {num_steps}")

        # Get set of dataset and model keys, ignoring any local tag index. That is, only the
        # global tag part is considered.
        datasets_keys = {LocalizedTag.from_tag(tag).global_tag for tag in datasets.keys()}
        models_keys = {LocalizedTag.from_tag(tag).global_tag for tag in models.keys()}
        if datasets_keys != models_keys:
            raise ValueError(
                f"datasets and models should contain the same keys. Got {datasets_keys} and"
                f" {models_keys} respectively."
            )

        if not datasets:
            raise ValueError("dicts of datasets and models must be populated.")

        if acquisition_rule is None:
            if datasets.keys() != {OBJECTIVE}:
                raise ValueError(
                    f"Default acquisition rule EfficientGlobalOptimization requires tag"
                    f" {OBJECTIVE!r}, got keys {datasets.keys()}"
                )

            acquisition_rule = EfficientGlobalOptimization[
                SearchSpaceType, TrainableProbabilisticModelType
            ]()

        history: list[
            FrozenRecord[StateType, TrainableProbabilisticModelType]
            | Record[StateType, TrainableProbabilisticModelType]
        ] = []
        history_state = acquisition_state
        query_plot_dfs: dict[int, pd.DataFrame] = {}
        observation_plot_dfs = observation_plot_init(datasets)

        summary_writer = logging.get_tensorboard_writer()
        if summary_writer:
            with summary_writer.as_default(step=0):
                write_summary_init(
                    self._observer,
                    self._search_space,
                    acquisition_rule,
                    datasets,
                    models,
                    num_steps,
                )

        for step in range(start_step + 1, num_steps + 1):
            logging.set_step_number(step)

            if early_stop_callback and early_stop_callback(datasets, models, acquisition_state):
                tf.print("Optimization terminated early", output_stream=absl.logging.INFO)
                break

            try:
                if track_state:
                    try:
                        # note that we use the state as it was at acquisition time, not the one
                        # modified in filter_datasets in preparation for the next acquisition,
                        # so we can restart the optimization correctly (and also for plotting)
                        if track_path is None:
                            datasets_copy = copy.deepcopy(datasets)
                            models_copy = copy.deepcopy(models)
                            history_state_copy = copy.deepcopy(history_state)
                            record = Record(datasets_copy, models_copy, history_state_copy)
                            history.append(record)
                        else:
                            track_path = Path(track_path)
                            record = Record(datasets, models, history_state)
                            file_name = OptimizationResult.step_filename(step, num_steps)
                            history.append(record.save(track_path / file_name))
                    except Exception as e:
                        raise NotImplementedError(
                            "Failed to save the optimization state. Some models do not support "
                            "deecopying or serialization and cannot be saved. "
                            "(This is particularly common for deep neural network models, though "
                            "some of the model wrappers accept a model closure as a workaround.) "
                            "For these models, the `track_state`` argument of the "
                            ":meth:`~trieste.bayesian_optimizer.BayesianOptimizer.optimize` method "
                            "should be set to `False`. This means that only the final model "
                            "will be available."
                        ) from e

                if step == 1:
                    # See explanation in AskTellOptimizer.__init__().
                    if isinstance(acquisition_rule, LocalDatasetsAcquisitionRule):
                        datasets = with_local_datasets(
                            datasets, acquisition_rule.num_local_datasets
                        )
                        acquisition_rule.initialize_subspaces(self._search_space)

                    filtered_datasets_or_callable: (
                        Mapping[Tag, Dataset] | State[StateType | None, Mapping[Tag, Dataset]]
                    ) = acquisition_rule.filter_datasets(models, datasets)
                    if callable(filtered_datasets_or_callable):
                        acquisition_state, filtered_datasets = filtered_datasets_or_callable(
                            acquisition_state
                        )
                    else:
                        filtered_datasets = filtered_datasets_or_callable

                    if fit_model and fit_initial_model:
                        with Timer() as initial_model_fitting_timer:
                            for tag, model in models.items():
                                # Prefer local dataset if available.
                                tags = [tag, LocalizedTag.from_tag(tag).global_tag]
                                _, dataset = get_value_for_tag(filtered_datasets, *tags)
                                assert dataset is not None
                                model.update(dataset)
                                optimize_model_and_save_result(model, dataset)
                        if summary_writer:
                            logging.set_step_number(0)
                            with summary_writer.as_default(step=0):
                                write_summary_initial_model_fit(
                                    datasets, models, initial_model_fitting_timer
                                )
                            logging.set_step_number(step)

                with Timer() as total_step_wallclock_timer:
                    with Timer() as query_point_generation_timer:
                        points_or_stateful = acquisition_rule.acquire(
                            self._search_space, models, datasets=filtered_datasets
                        )
                        if callable(points_or_stateful):
                            acquisition_state, query_points = points_or_stateful(acquisition_state)
                            history_state = acquisition_state
                        else:
                            query_points = points_or_stateful

                    observer = self._observer
                    # If query_points are rank 3, then use a batched observer.
                    if tf.rank(query_points) == 3:
                        observer = mk_batch_observer(observer)
                    observer_output = observer(query_points)

                    tagged_output = (
                        observer_output
                        if isinstance(observer_output, Mapping)
                        else {OBJECTIVE: observer_output}
                    )

                    for tag, new_dataset in tagged_output.items():
                        datasets[tag] += new_dataset

                    filtered_datasets_or_callable = acquisition_rule.filter_datasets(
                        models, datasets
                    )
                    if callable(filtered_datasets_or_callable):
                        acquisition_state, filtered_datasets = filtered_datasets_or_callable(
                            acquisition_state
                        )
                    else:
                        filtered_datasets = filtered_datasets_or_callable

                    with Timer() as model_fitting_timer:
                        if fit_model:
                            for tag, model in models.items():
                                # Always use the matching dataset to the model. If the model is
                                # local, then the dataset should be too by this stage.
                                dataset = filtered_datasets[tag]
                                model.update(dataset)
                                optimize_model_and_save_result(model, dataset)

                if summary_writer:
                    with summary_writer.as_default(step=step):
                        write_summary_observations(
                            datasets,
                            models,
                            tagged_output,
                            model_fitting_timer,
                            observation_plot_dfs,
                        )
                        write_summary_query_points(
                            datasets,
                            models,
                            self._search_space,
                            query_points,
                            query_point_generation_timer,
                            query_plot_dfs,
                        )
                        logging.scalar("wallclock/step", total_step_wallclock_timer.time)

            except Exception as error:  # pylint: disable=broad-except
                tf.print(
                    f"\nOptimization failed at step {step}, encountered error with traceback:"
                    f"\n{traceback.format_exc()}"
                    f"\nTerminating optimization and returning the optimization history. You may "
                    f"be able to use the history to restart the process from a previous successful "
                    f"optimization step.\n",
                    output_stream=absl.logging.ERROR,
                )
                if isinstance(error, MemoryError):
                    tf.print(
                        "\nOne possible cause of memory errors is trying to evaluate acquisition "
                        "\nfunctions over large datasets, e.g. when initializing optimizers. "
                        "\nYou may be able to word around this by splitting up the evaluation "
                        "\nusing split_acquisition_function or split_acquisition_function_calls.",
                        output_stream=absl.logging.ERROR,
                    )
                result = OptimizationResult(Err(error), history)
                if track_state and track_path is not None:
                    result.save_result(Path(track_path) / OptimizationResult.RESULTS_FILENAME)
                return result

        tf.print("Optimization completed without errors", output_stream=absl.logging.INFO)

        record = Record(datasets, models, history_state)
        result = OptimizationResult(Ok(record), history)
        if track_state and track_path is not None:
            result.save_result(Path(track_path) / OptimizationResult.RESULTS_FILENAME)
        return result

    def continue_optimization(
        self,
        num_steps: int,
        optimization_result: OptimizationResult[StateType, TrainableProbabilisticModelType],
        *args: Any,
        **kwargs: Any,
    ) -> OptimizationResult[StateType, TrainableProbabilisticModelType]:
        """
        Continue a previous optimization that either failed, was terminated early, or which
        you simply wish to run for more steps.

        :param num_steps: The total number of optimization steps, including any that have already
            been run.
        :param optimization_result: The optimization result from which to extract the datasets,
            models and acquisition state. If the result was successful then the final result is
            used; otherwise the last record in the history is used. The size of the history
            is used to determine how many more steps are required.
        :param args: Any more positional arguments to pass on to optimize.
        :param kwargs: Any more keyword arguments to pass on to optimize.
        :return: An :class:`OptimizationResult`. The history will contain both the history from
            `optimization_result` (including the `final_result` if that was successful) and
            any new records.
        """
        history: list[
            Record[StateType, TrainableProbabilisticModelType]
            | FrozenRecord[StateType, TrainableProbabilisticModelType]
        ] = []
        history.extend(optimization_result.history)
        if optimization_result.final_result.is_ok:
            history.append(optimization_result.final_result.unwrap())
        if not history:
            raise ValueError("Cannot continue from empty optimization result")

        result = self.optimize(  # type: ignore[call-overload]
            num_steps,
            history[-1].datasets,
            history[-1].models,
            *args,
            acquisition_state=history[-1].acquisition_state,
            **kwargs,
            start_step=len(history) - 1,
        )
        result.history[:1] = history
        return result


def write_summary_init(
    observer: Observer,
    search_space: SearchSpace,
    acquisition_rule: AcquisitionRule[
        TensorType | State[StateType | None, TensorType],
        SearchSpaceType,
        TrainableProbabilisticModelType,
    ],
    datasets: Mapping[Tag, Dataset],
    models: Mapping[Tag, TrainableProbabilisticModel],
    num_steps: int,
) -> None:
    """Write initial BO loop TensorBoard summary."""
    devices = tf.config.list_logical_devices()
    logging.text(
        "metadata",
        f"Observer: `{observer}`\n\n"
        f"Number of steps: `{num_steps}`\n\n"
        f"Number of initial points: "
        f"`{dict((k, len(v)) for k, v in datasets.items())}`\n\n"
        f"Search Space: `{search_space}`\n\n"
        f"Acquisition rule:\n\n    {acquisition_rule}\n\n"
        f"Models:\n\n    {models}\n\n"
        f"Available devices: `{dict(Counter(d.device_type for d in devices))}`",
    )


def write_summary_initial_model_fit(
    datasets: Mapping[Tag, Dataset],
    models: Mapping[Tag, ProbabilisticModel],
    model_fitting_timer: Timer,
) -> None:
    """Write TensorBoard summary for the model fitting to the initial data."""
    for tag, model in models.items():
        with tf.name_scope(f"{tag}.model"):
            # Prefer local dataset if available.
            tags = [tag, LocalizedTag.from_tag(tag).global_tag]
            _, dataset = get_value_for_tag(datasets, *tags)
            assert dataset is not None
            model.log(dataset)
    logging.scalar(
        "wallclock/model_fitting",
        model_fitting_timer.time,
    )


def observation_plot_init(
    datasets: Mapping[Tag, Dataset],
) -> dict[Tag, pd.DataFrame]:
    """Initialise query point pairplot dataframes with initial observations.
    Also logs warnings if pairplot dependencies are not installed."""
    observation_plot_dfs: dict[Tag, pd.DataFrame] = {}
    if logging.get_tensorboard_writer():
        seaborn_warning = False
        if logging.include_summary("query_points/_pairplot") and not (pd and sns):
            seaborn_warning = True
        for tag in datasets:
            if logging.include_summary(f"{tag}.observations/_pairplot"):
                output_dim = tf.shape(datasets[tag].observations)[-1]
                if output_dim >= 2:
                    if not (pd and sns):
                        seaborn_warning = True
                    else:
                        columns = [f"x{i}" for i in range(output_dim)]
                        observation_plot_dfs[tag] = pd.DataFrame(
                            datasets[tag].observations, columns=columns
                        ).applymap(float)
                        observation_plot_dfs[tag]["observations"] = "initial"

        if seaborn_warning:
            tf.print(
                "\nPairplot TensorBoard summaries require seaborn to be installed."
                "\nOne way to do this is to install 'trieste[plotting]'.",
                output_stream=absl.logging.INFO,
            )
    return observation_plot_dfs


def write_summary_observations(
    datasets: Mapping[Tag, Dataset],
    models: Mapping[Tag, ProbabilisticModel],
    tagged_output: Mapping[Tag, TensorType],
    model_fitting_timer: Timer,
    observation_plot_dfs: MutableMapping[Tag, pd.DataFrame],
) -> None:
    """Write TensorBoard summary for the current step observations."""
    for tag in models:
        if tag not in tagged_output:
            continue

        with tf.name_scope(f"{tag}.model"):
            models[tag].log(datasets[tag])

        output_dim = tf.shape(tagged_output[tag].observations)[-1]
        for i in tf.range(output_dim):
            suffix = f"[{i}]" if output_dim > 1 else ""
            if tf.size(tagged_output[tag].observations) > 0:
                logging.histogram(
                    f"{tag}.observation{suffix}/new_observations",
                    tagged_output[tag].observations[..., i],
                )
                logging.scalar(
                    f"{tag}.observation{suffix}/best_new_observation",
                    np.min(tagged_output[tag].observations[..., i]),
                )
            if tf.size(datasets[tag].observations) > 0:
                logging.scalar(
                    f"{tag}.observation{suffix}/best_overall",
                    np.min(datasets[tag].observations[..., i]),
                )

        if logging.include_summary(f"{tag}.observations/_pairplot") and (
            pd and sns and output_dim >= 2
        ):
            columns = [f"x{i}" for i in range(output_dim)]
            observation_new_df = pd.DataFrame(
                tagged_output[tag].observations, columns=columns
            ).applymap(float)
            observation_new_df["observations"] = "new"
            observation_plot_df = pd.concat(
                (observation_plot_dfs.get(tag), observation_new_df),
                copy=False,
                ignore_index=True,
            )

            hue_order = ["initial", "old", "new"]
            palette = {"initial": "tab:green", "old": "tab:green", "new": "tab:orange"}
            markers = {"initial": "X", "old": "o", "new": "o"}

            # assume that any OBJECTIVE- or single-tagged multi-output dataset => multi-objective
            # more complex scenarios (e.g. constrained data) need to be plotted by the acq function
            if len(datasets) > 1 and tag != OBJECTIVE:
                observation_plot_df["observation type"] = observation_plot_df.apply(
                    lambda x: x["observations"],
                    axis=1,
                )
            else:
                observation_plot_df["pareto"] = non_dominated(datasets[tag].observations)[1]
                observation_plot_df["observation type"] = observation_plot_df.apply(
                    lambda x: x["observations"] + x["pareto"] * " (non-dominated)",
                    axis=1,
                )
                hue_order += [hue + " (non-dominated)" for hue in hue_order]
                palette.update(
                    {
                        "initial (non-dominated)": "tab:purple",
                        "old (non-dominated)": "tab:purple",
                        "new (non-dominated)": "tab:red",
                    }
                )
                markers.update(
                    {
                        "initial (non-dominated)": "X",
                        "old (non-dominated)": "o",
                        "new (non-dominated)": "o",
                    }
                )

            pairplot = sns.pairplot(
                observation_plot_df,
                vars=columns,
                hue="observation type",
                hue_order=hue_order,
                palette=palette,
                markers=markers,
            )
            logging.pyplot(f"{tag}.observations/_pairplot", pairplot.fig)
            observation_plot_df.loc[
                observation_plot_df["observations"] == "new", "observations"
            ] = "old"
            observation_plot_dfs[tag] = observation_plot_df

    logging.scalar(
        "wallclock/model_fitting",
        model_fitting_timer.time,
    )


def write_summary_query_points(
    datasets: Mapping[Tag, Dataset],
    models: Mapping[Tag, ProbabilisticModel],
    search_space: SearchSpace,
    query_points: TensorType,
    query_point_generation_timer: Timer,
    query_plot_dfs: MutableMapping[int, pd.DataFrame],
) -> None:
    """Write TensorBoard summary for the current step query points."""

    if tf.rank(query_points) == 2:
        for i in tf.range(tf.shape(query_points)[1]):
            if len(query_points) == 1:
                logging.scalar(f"query_point/[{i}]", float(query_points[0, i]))
            else:
                logging.histogram(f"query_points/[{i}]", query_points[:, i])
        logging.histogram("query_points/euclidean_distances", lambda: pdist(query_points))

    if pd and sns and logging.include_summary("query_points/_pairplot"):
        columns = [f"x{i}" for i in range(tf.shape(query_points)[1])]
        qp_preds = query_points
        for tag in datasets:
            pred = models[tag].predict(query_points)[0]
            qp_preds = tf.concat([qp_preds, tf.cast(pred, query_points.dtype)], 1)
            output_dim = tf.shape(pred)[-1]
            for i in range(output_dim):
                columns.append(f"{tag}{i if (output_dim > 1) else ''} predicted")
        query_new_df = pd.DataFrame(qp_preds, columns=columns).applymap(float)
        query_new_df["query points"] = "new"
        query_plot_df = pd.concat(
            (query_plot_dfs.get(0), query_new_df), copy=False, ignore_index=True
        )
        pairplot = sns.pairplot(
            query_plot_df, hue="query points", hue_order=["old", "new"], height=2.25
        )
        padding = 0.025 * (search_space.upper - search_space.lower)
        upper_limits = search_space.upper + padding
        lower_limits = search_space.lower - padding
        for i in range(search_space.dimension):
            pairplot.axes[0, i].set_xlim((lower_limits[i], upper_limits[i]))
            pairplot.axes[i, 0].set_ylim((lower_limits[i], upper_limits[i]))
        logging.pyplot("query_points/_pairplot", pairplot.fig)
        query_plot_df["query points"] = "old"
        query_plot_dfs[0] = query_plot_df

    logging.scalar(
        "wallclock/query_point_generation",
        query_point_generation_timer.time,
    )


def stop_at_minimum(
    minimum: Optional[tf.Tensor] = None,
    minimizers: Optional[tf.Tensor] = None,
    minimum_atol: float = 0,
    minimum_rtol: float = 0.05,
    minimizers_atol: float = 0,
    minimizers_rtol: float = 0.05,
    objective_tag: Tag = OBJECTIVE,
    minimum_step_number: Optional[int] = None,
) -> EarlyStopCallback[TrainableProbabilisticModel, object]:
    """
    Generate an early stop function that terminates a BO loop when it gets close enough to the
    given objective minimum and/or minimizer points.

    :param minimum: Optional minimum to stop at, with shape [1].
    :param minimizers: Optional minimizer points to stop at, with shape [N, D].
    :param minimum_atol: Absolute tolerance for minimum.
    :param minimum_rtol: Relative tolerance for minimum.
    :param minimizers_atol: Absolute tolerance for minimizer point.
    :param minimizers_rtol: Relative tolerance for minimizer point.
    :param objective_tag: The tag for the objective data.
    :param minimum_step_number: Minimum step number to stop at.
    :return: An early stop function that terminates if we get close enough to both the minimum
        and any of the minimizer points.
    """

    def early_stop_callback(
        datasets: Mapping[Tag, Dataset],
        _models: Mapping[Tag, TrainableProbabilisticModel],
        _acquisition_state: object,
    ) -> bool:
        if minimum_step_number is not None and logging.get_step_number() < minimum_step_number:
            return False
        dataset = datasets[objective_tag]
        arg_min_idx = tf.squeeze(tf.argmin(dataset.observations, axis=0))
        if minimum is not None:
            best_y = dataset.observations[arg_min_idx]
            close_y = np.isclose(best_y, minimum, atol=minimum_atol, rtol=minimum_rtol)
            if not tf.reduce_all(close_y):
                return False
        if minimizers is not None:
            best_x = dataset.query_points[arg_min_idx]
            close_x = np.isclose(best_x, minimizers, atol=minimizers_atol, rtol=minimizers_rtol)
            if not tf.reduce_any(tf.reduce_all(close_x, axis=-1), axis=0):
                return False
        return True

    return early_stop_callback



# Copyright 2022 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import functools
from typing import Dict, Mapping, Optional, Sequence, Tuple, Union

import tensorflow as tf
from check_shapes import check_shapes

from ..data import Dataset
from ..models import ProbabilisticModelType
from ..observer import OBJECTIVE
from ..space import SearchSpaceType
from ..types import Tag, TensorType
from ..utils.misc import LocalizedTag
from .interface import AcquisitionFunction
from .optimizer import AcquisitionOptimizer


def split_acquisition_function(
    fn: AcquisitionFunction,
    split_size: int,
) -> AcquisitionFunction:
    """
    A wrapper around an :const:`AcquisitionFunction` to split its input into batches.
    Splits `x` into batches along the first dimension, calls `fn` on each batch, and then stitches
    the results back together, so that it looks like `fn` was called with all of `x` in one batch.
    :param fn: Acquisition function to split.
    :param split_size: Call fn with tensors of at most this size.
    :return: Split acquisition function.
    """
    if split_size <= 0:
        raise ValueError(f"split_size must be positive, got {split_size}")

    @functools.wraps(fn)
    def wrapper(x: TensorType) -> TensorType:
        x = tf.convert_to_tensor(x)

        # this currently assumes leading dimension of x is the split dimension.
        length = x.shape[0]
        if length == 0:
            return fn(x)

        # Use int64 to calculate the input tensor size, otherwise we can overflow for large tensors.
        elements_per_block = tf.size(x, out_type=tf.int64) / length
        blocks_per_batch = tf.cast(tf.math.ceil(split_size / elements_per_block), tf.int32)

        num_batches = tf.cast(tf.math.ceil(length / blocks_per_batch) - 1, tf.int32)
        batch_sizes = tf.concat(
            [
                tf.ones(num_batches, tf.int32) * blocks_per_batch,
                [length - num_batches * blocks_per_batch],
            ],
            axis=0,
        )

        if batch_sizes.shape[0] <= 1:
            return fn(x)

        batch_inputs = tf.split(x, batch_sizes)

        batch_outputs = []
        for batch_input in batch_inputs:
            output = fn(batch_input)
            batch_outputs.append(output)

        return tf.concat(batch_outputs, axis=0)

    return wrapper


def split_acquisition_function_calls(
    optimizer: AcquisitionOptimizer[SearchSpaceType],
    split_size: int,
) -> AcquisitionOptimizer[SearchSpaceType]:
    """
    A wrapper around our :const:`AcquisitionOptimizer`s. This class wraps a
    :const:`AcquisitionOptimizer` so that evaluations of the acquisition functions
    are split into batches on the first dimension and then stitched back together.
    This can be useful to reduce memory usage when evaluating functions over large spaces.

    :param optimizer: An optimizer that returns batches of points with shape [V, ...].
    :param split_size: The desired maximum number of points in acquisition function evaluations.
    :return: An :const:`AcquisitionOptimizer` that still returns points with the shape [V, ...]
        but evaluates at most split_size points at a time.
    """
    if split_size <= 0:
        raise ValueError(f"split_size must be positive, got {split_size}")

    def split_optimizer(
        search_space: SearchSpaceType,
        f: Union[AcquisitionFunction, Tuple[AcquisitionFunction, int]],
    ) -> TensorType:
        af, n = f if isinstance(f, tuple) else (f, 1)
        taf = split_acquisition_function(af, split_size)
        return optimizer(search_space, (taf, n) if isinstance(f, tuple) else taf)

    return split_optimizer


def select_nth_output(x: TensorType, output_dim: int = 0) -> TensorType:
    """
    A utility function for trajectory sampler-related acquisition functions which selects the `n`th
    output as the trajectory to be used, with `n` specified by ``output_dim``. Defaults to the first
    output.

    :param x: Input with shape [..., B, L], where L is the number of outputs of the model.
    :param output_dim: Dimension of the output to be selected. Defaults to the first output.
    :return: TensorType with shape [..., B], where the output_dim dimension has been selected to
        reduce the input.
    """
    return x[..., output_dim]


def get_local_dataset(local_space: SearchSpaceType, dataset: Dataset) -> Dataset:
    """
    A utility function that takes in a dataset and returns the entries lying
    within a given search space.

    :param local_space: A search space.
    :param dataset: A Dataset.
    :return: A Dataset containing entries only in the local_space.
    """
    if tf.shape(dataset.query_points)[1] != local_space.dimension:
        raise ValueError("Dataset and search space must have equal dimensions")

    is_in_region_mask = local_space.contains(dataset.query_points)
    local_dataset = Dataset(
        query_points=tf.boolean_mask(dataset.query_points, is_in_region_mask),
        observations=tf.boolean_mask(dataset.observations, is_in_region_mask),
    )
    return local_dataset


def copy_to_local_models(
    global_model: ProbabilisticModelType,
    num_local_models: int,
    key: Tag = OBJECTIVE,
) -> Mapping[Tag, ProbabilisticModelType]:
    """
    Helper method to copy a global model to local models.

    :param global_model: The global model.
    :param num_local_models: The number of local models to create.
    :param key: The tag prefix for the local models.
    :return: A mapping of the local models.
    """
    return {LocalizedTag(key, i): copy.deepcopy(global_model) for i in range(num_local_models)}


def with_local_datasets(
    datasets: Mapping[Tag, Dataset],
    num_local_datasets: int,
    local_dataset_indices: Optional[Sequence[TensorType]] = None,
) -> Dict[Tag, Dataset]:
    """
    Helper method to add local datasets if they do not already exist, by copying global datasets
    or a subset thereof.

    :param datasets: The original datasets.
    :param num_local_datasets: The number of local datasets to add per global tag.
    :param local_dataset_indices: Optional sequence of indices, indicating which parts of
        the global datasets should be copied. If None then the entire datasets are copied.
    :return: The updated mapping of datasets.
    """
    if local_dataset_indices is not None and len(local_dataset_indices) != num_local_datasets:
        raise ValueError(
            f"local_dataset_indices should have {num_local_datasets} entries, "
            f"has {len(local_dataset_indices)}"
        )

    updated_datasets = {}
    for tag in datasets:
        updated_datasets[tag] = datasets[tag]
        ltag = LocalizedTag.from_tag(tag)
        if not ltag.is_local:
            for i in range(num_local_datasets):
                target_ltag = LocalizedTag(ltag.global_tag, i)
                if target_ltag not in datasets:
                    if local_dataset_indices is None:
                        updated_datasets[target_ltag] = datasets[tag]
                    else:
                        # TODO: use sparse tensors instead
                        updated_datasets[target_ltag] = Dataset(
                            query_points=tf.gather(
                                datasets[tag].query_points, local_dataset_indices[i]
                            ),
                            observations=tf.gather(
                                datasets[tag].observations, local_dataset_indices[i]
                            ),
                        )

    return updated_datasets


@check_shapes(
    "points: [n_points, ...]",
    "return: [n_points]",
)
def get_unique_points_mask(points: TensorType, tolerance: float = 1e-6) -> TensorType:
    """Find the boolean mask of unique points in a tensor, within a given tolerance.

    Users can get the actual points with:

        mask = get_unique_points_mask(points, tolerance)
        unique_points = tf.boolean_mask(points, mask)

    Note: this uses a greedy parallel set covering algorithm, so isn't guaranteed to find the
    smallest possible set of unique points: e.g. for the points [[1],[2],[3]] with tolerance 1,
    it returns [True, False True] rather than [False, True, False].

    :param points: A tensor of points, with the first dimension being the number of points.
    :param tolerance: The tolerance up to which points are considered equal.
    :return: A boolean mask for the unique points.
    """

    # Calculate the pairwise "equality" between points
    pairwise_distances = tf.math.reduce_euclidean_norm(
        tf.expand_dims(points, 1) - tf.expand_dims(points, 0), axis=-1
    )
    pairwise_equal = pairwise_distances <= tolerance

    # Replace the upper triangle with False. That way, any row that is all False
    # is not equal to any point before it, so can be safely included in the output.
    upper_triangle = tf.linalg.band_part(tf.ones_like(pairwise_equal), 0, -1)
    triangle_equal = tf.logical_and(pairwise_equal, ~upper_triangle)

    # The converse is not true, however: a row that is not all False might still need
    # to be included if the only Trues in it correspond to other points that weren't selected.
    # For example, [1,2,3,4] with tolerance 1 should select not just 1 but also 3, because
    # even though 3 is equal to 2, it isn't equal to 1. We therefore keep track of which points
    # remain candidates for uniqueness, and repeat the triangle-based calculation with
    # those until there are no candidates left.
    candidate_mask = tf.ones(shape=tf.shape(points)[:1], dtype=tf.bool)
    unique_mask = tf.zeros(shape=tf.shape(points)[:1], dtype=tf.bool)
    while tf.reduce_any(candidate_mask):
        new_points = ~tf.reduce_any(triangle_equal, axis=1)
        unique_mask = tf.logical_or(unique_mask, new_points)
        equal_to_new_points = tf.reduce_any(pairwise_equal[new_points], axis=0)
        candidate_mask = tf.logical_and(candidate_mask, ~equal_to_new_points)
        # update triangle_equal to ignore non-candidate points
        triangle_equal = tf.where(candidate_mask, triangle_equal, tf.zeros_like(triangle_equal))
        triangle_equal = tf.where(
            tf.expand_dims(candidate_mask, 1), triangle_equal, tf.ones_like(triangle_equal)
        )

    return unique_mask



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Optional

import tensorflow as tf
import tensorflow_probability as tfp

from ...data import Dataset
from ...types import TensorType


def get_tensor_spec_from_data(dataset: Dataset) -> tuple[tf.TensorSpec, tf.TensorSpec]:
    r"""
    Extract tensor specifications for inputs and outputs of neural network models, based on the
    dataset. This utility faciliates constructing neural networks, providing the required
    dimensions for the input and the output of the network. For example

    >>> data = Dataset(
    ...     tf.constant([[0.1, 0.2], [0.3, 0.4]]),
    ...     tf.constant([[0.5], [0.7]])
    ... )
    >>> input_spec, output_spec = get_tensor_spec_from_data(data)
    >>> input_spec
    TensorSpec(shape=(2,), dtype=tf.float32, name='query_points')
    >>> output_spec
    TensorSpec(shape=(1,), dtype=tf.float32, name='observations')

    :param dataset: A dataset with ``query_points`` and ``observations`` tensors.
    :return: Tensor specification objects for the ``query_points`` and ``observations`` tensors.
    :raise ValueError: If the dataset is not an instance of :class:`~trieste.data.Dataset`.
    """
    if not isinstance(dataset, Dataset):
        raise ValueError(
            f"This function works only on trieste.data.Dataset objects, however got"
            f"{type(dataset)} which is incompatible."
        )
    input_tensor_spec = tf.TensorSpec(
        shape=(dataset.query_points.shape[1:]),
        dtype=dataset.query_points.dtype,
        name="query_points",
    )
    output_tensor_spec = tf.TensorSpec(
        shape=(dataset.observations.shape[1:]),
        dtype=dataset.observations.dtype,
        name="observations",
    )
    return input_tensor_spec, output_tensor_spec


def sample_with_replacement(dataset: Dataset) -> Dataset:
    """
    Create a new ``dataset`` with data sampled with replacement. This
    function is useful for creating bootstrap samples of data for training ensembles.

    :param dataset: The data that should be sampled.
    :return: A (new) ``dataset`` with sampled data.
    :raise ValueError (or InvalidArgumentError): If the dataset is not an instance of
        :class:`~trieste.data.Dataset` or it is empty.
    """
    if not isinstance(dataset, Dataset):
        raise ValueError(
            f"This function works only on trieste.data.Dataset objects, however got"
            f"{type(dataset)} which is incompatible."
        )
    tf.debugging.assert_positive(len(dataset), message="Dataset must not be empty.")

    n_rows = dataset.observations.shape[0]

    index_tensor = tf.random.uniform((n_rows,), maxval=n_rows, dtype=tf.dtypes.int32)

    observations = tf.gather(dataset.observations, index_tensor, axis=0)
    query_points = tf.gather(dataset.query_points, index_tensor, axis=0)

    return Dataset(query_points=query_points, observations=observations)


def sample_model_index(
    size: TensorType,
    num_samples: TensorType,
    seed: Optional[int] = None,
) -> TensorType:
    """
    Returns samples of indices of individual models in the ensemble.

    If ``num_samples`` is smaller or equal to ``size`` (i.e. the ensemble size) indices are sampled
    without replacement. When ``num_samples`` is larger than ``size`` then until ``size`` is reached
    we sample without replacement, while after that we sample with replacement. The rationale of
    this mixed scheme is that typically one wants to exhaust all networks and then resample them
    only if required.

    :param size: The maximum index, effectively the number of models in the ensemble.
    :param num_samples: The number of samples to take.
    :param seed: Optional RNG seed.
    :return: A tensor with indices.
    """
    shuffle_indices = tf.random.shuffle(tf.range(size), seed=seed)
    if num_samples > size:
        random_indices = tf.random.uniform(
            shape=(tf.cast(num_samples - size, tf.int32),),
            maxval=size,
            dtype=tf.int32,
            seed=seed,
        )
        indices = tf.concat([shuffle_indices, random_indices], 0)
    else:
        indices = shuffle_indices[:num_samples]

    return indices


def negative_log_likelihood(
    y_true: TensorType, y_pred: tfp.distributions.Distribution
) -> TensorType:
    """
    Maximum likelihood objective function for training neural networks.

    :param y_true: The output variable values.
    :param y_pred: The output layer of the model. It has to be a probabilistic neural network
        with a distribution as a final layer.
    :return: Negative log likelihood values.
    """
    return -y_pred.log_prob(y_true)



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains acquisition function builders for continuous Thompson sampling.
"""
from __future__ import annotations

from typing import Any, Callable, Optional, Type

import tensorflow as tf

from ...data import Dataset
from ...models.interfaces import HasTrajectorySampler, TrajectoryFunction, TrajectoryFunctionClass
from ...types import TensorType
from ..interface import SingleModelGreedyAcquisitionBuilder, SingleModelVectorizedAcquisitionBuilder
from ..utils import select_nth_output


class GreedyContinuousThompsonSampling(SingleModelGreedyAcquisitionBuilder[HasTrajectorySampler]):
    r"""

    Acquisition function builder for performing greedy continuous Thompson sampling. This builder
    return acquisition functions that are the negatives of approximate samples from the
    given :class:`ProbabilisticModel`, as provided by the model's :meth:`get_trajectory`
    method. A set of such samples are to be maximized in a sequential greedy manner to provide
    the next recommended query points. Note that we actually return
    the negative of the trajectory, so that our acquisition optimizers (which are
    all maximizers) can be used to extract the minimisers of trajectories.


    For more details about trajectory-based Thompson sampling see :cite:`hernandez2017parallel` and
    :cite:`wilson2020efficiently`.
    """

    def __init__(self, select_output: Callable[[TensorType], TensorType] = select_nth_output):
        """
        :param select_output: A method that returns the desired trajectory from a trajectory
            sampler with shape [..., B], where B is a batch dimension. Defaults to the
            :func:~`trieste.acquisition.utils.select_nth_output` function with output dimension 0.
        """
        self._select_output = select_output

    def __repr__(self) -> str:
        """"""
        return f"GreedyContinuousThompsonSampling({self._select_output!r})"

    def prepare_acquisition_function(
        self,
        model: HasTrajectorySampler,
        dataset: Optional[Dataset] = None,
        pending_points: Optional[TensorType] = None,
    ) -> TrajectoryFunction:
        """
        :param model: The model.
        :param dataset: The data from the observer (not used).
        :param pending_points: The points already in the current batch (not used).
        :return: A negated trajectory sampled from the model.
        """
        if not isinstance(model, HasTrajectorySampler):
            raise ValueError(
                f"Thompson sampling from trajectory only supports models with a trajectory_sampler "
                f"method; received {model!r}"
            )

        self._trajectory_sampler = model.trajectory_sampler()
        function = self._trajectory_sampler.get_trajectory()
        return negate_trajectory_function(function, self._select_output)

    def update_acquisition_function(
        self,
        function: TrajectoryFunction,
        model: HasTrajectorySampler,
        dataset: Optional[Dataset] = None,
        pending_points: Optional[TensorType] = None,
        new_optimization_step: bool = True,
    ) -> TrajectoryFunction:
        """
        :param function: The trajectory function to update.
        :param model: The model.
        :param dataset: The data from the observer (not used).
        :param pending_points: The points already in the current batch (not used).
        :param new_optimization_step: Indicates whether this call to update_acquisition_function
            is to start of a new optimization step, of to continue collecting batch of points
            for the current step. Defaults to ``True``.
        :return: A new trajectory sampled from the model.
        """

        if new_optimization_step:  # update sampler and resample trajectory
            new_function = self._trajectory_sampler.update_trajectory(function)
        else:  # just resample trajectory but without updating sampler
            new_function = self._trajectory_sampler.resample_trajectory(function)

        if new_function is not function:
            function = negate_trajectory_function(new_function, self._select_output)

        return function


class ParallelContinuousThompsonSampling(
    SingleModelVectorizedAcquisitionBuilder[HasTrajectorySampler]
):
    r"""
    Acquisition function builder for performing parallel continuous Thompson sampling.

    This builder provides broadly the same behavior as our :class:`GreedyContinuousThompsonSampler`
    however optimizes trajectory samples in parallel rather than sequentially.
    Consequently, :class:`ParallelContinuousThompsonSampling` can choose query points faster
    than  :class:`GreedyContinuousThompsonSampler` however it has much larger memory usage.

    For a convenient way to control the total memory usage of this acquisition function, see
    our :const:`split_acquisition_function_calls` wrapper.
    """

    def __init__(self, select_output: Callable[[TensorType], TensorType] = select_nth_output):
        """
        :param select_output: A method that returns the desired trajectory from a trajectory
            sampler with shape [..., B], where B is a batch dimension. Defaults to the
            :func:~`trieste.acquisition.utils.select_nth_output` function with output dimension 0.
        """
        self._select_output = select_output

    def __repr__(self) -> str:
        """"""
        return f"ParallelContinuousThompsonSampling({self._select_output!r})"

    def prepare_acquisition_function(
        self,
        model: HasTrajectorySampler,
        dataset: Optional[Dataset] = None,
    ) -> TrajectoryFunction:
        """
        :param model: The model.
        :param dataset: The data from the observer (not used).
        :return: A negated trajectory sampled from the model.
        """
        if not isinstance(model, HasTrajectorySampler):
            raise ValueError(
                f"Thompson sampling from trajectory only supports models with a trajectory_sampler "
                f"method; received {model!r}"
            )

        self._trajectory_sampler = model.trajectory_sampler()
        self._trajectory = self._trajectory_sampler.get_trajectory()
        self._negated_trajectory = negate_trajectory_function(self._trajectory, self._select_output)
        return self._negated_trajectory

    def update_acquisition_function(
        self,
        function: TrajectoryFunction,
        model: HasTrajectorySampler,
        dataset: Optional[Dataset] = None,
    ) -> TrajectoryFunction:
        """
        :param function: The trajectory function to update.
        :param model: The model.
        :param dataset: The data from the observer (not used).
        :return: A new trajectory sampled from the model.
        """
        if function is not self._negated_trajectory:
            raise ValueError("Wrong trajectory function passed into update_acquisition_function")

        new_function = self._trajectory_sampler.update_trajectory(self._trajectory)

        if new_function is not self._trajectory:  # need to negate again if not modified in place
            self._trajectory = new_function
            self._negated_trajectory = negate_trajectory_function(new_function, self._select_output)

        return self._negated_trajectory


class _DummyTrajectoryFunctionClass(TrajectoryFunctionClass):
    # dummy trajectory function class used while pickling NegatedTrajectory
    def __call__(self, x: TensorType) -> TensorType:
        return x


def negate_trajectory_function(
    function: TrajectoryFunction,
    select_output: Optional[Callable[[TensorType], TensorType]] = None,
    function_type: Optional[Type[TrajectoryFunction]] = None,
) -> TrajectoryFunction:
    """
    Return the negative of trajectories and select the output to form the acquisition function, so
    that our acquisition optimizers (which are all maximizers) can be used to extract the minimizers
    of trajectories.

    We negate the trajectory function object's call method, as it may have e.g. update and resample
    methods, and select the output we wish to use.
    """
    if isinstance(function, TrajectoryFunctionClass):

        class NegatedTrajectory(function_type or type(function)):  # type: ignore[misc]
            @tf.function
            def __call__(self, x: TensorType) -> TensorType:
                if select_output is not None:
                    return -1.0 * select_output(super().__call__(x))
                else:
                    return -1.0 * super().__call__(x)

            def __reduce__(
                self,
            ) -> tuple[
                Callable[..., TrajectoryFunction],
                tuple[
                    TrajectoryFunction,
                    Optional[Callable[[TensorType], TensorType]],
                    Optional[Type[TrajectoryFunction]],
                ],
                dict[str, Any],
            ]:
                # make this pickleable
                state = (
                    self.__getstate__() if hasattr(self, "__getstate__") else self.__dict__.copy()
                )
                return (
                    negate_trajectory_function,
                    (_DummyTrajectoryFunctionClass(), select_output, self.__class__.__base__),
                    state,
                )

        function.__class__ = NegatedTrajectory

        return function

    else:

        @tf.function
        def negated_trajectory(x: TensorType) -> TensorType:
            if select_output is not None:
                return -1.0 * select_output(function(x))
            else:
                return -1.0 * function(x)

        return negated_trajectory



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This package contains examples of popular objective functions used in (Bayesian) optimization.
"""

from . import multi_objectives, utils
from .multi_objectives import DTLZ1, DTLZ2, VLMOP2, MultiObjectiveTestProblem
from .multifidelity_objectives import (
    Linear2Fidelity,
    Linear3Fidelity,
    Linear5Fidelity,
    SingleObjectiveMultifidelityTestProblem,
)
from .single_objectives import (
    Ackley5,
    Branin,
    ConstrainedScaledBranin,
    GramacyLee,
    Hartmann3,
    Hartmann6,
    Levy8,
    LogarithmicGoldsteinPrice,
    Michalewicz2,
    Michalewicz5,
    Michalewicz10,
    ObjectiveTestProblem,
    Rosenbrock4,
    ScaledBranin,
    Shekel4,
    SimpleQuadratic,
    SingleObjectiveTestProblem,
    Trid10,
)
from .utils import mk_multi_observer, mk_observer
